
import csv
import os
import pandas as pd
import sys

from dataclasses import dataclass
from typing import Dict, List, Set

RESULTS_DIR = "processed-results" + os.sep
Z3 = "z3"
SYMPARIKH = "sym_parikh"
RESULTS_SUFFIX = "_result"
TIME = "time"
TIME_SUFFIX = "_" + TIME
COUNT = "count"
COUNT_SUFFIX = "_" + COUNT

INCOMPATIBLE = "Incompatible"
EASY_SAT = "Easy (sat)"
EASY_UNSAT = "Easy (unsat)"
ERROR = "Error"
NO_RESULT = ""
SAT = "sat"
UNSAT = "unsat"

TIMES_DAT_SUFFIX = "_times.dat"

MARK_REPEAT = "10"
STYLES_DEFINITIONS = """
    \\tikzset{decoone/.style={
            orange,
            decoration={
                markings,
                mark = between positions 0 and 1 step 5mm with {
                    \\node[cross out,draw=orange,inner sep=0mm,minimum size=1mm]{};
                },
            },
            postaction={decorate}
        }
    }
    \\tikzset{decotwo/.style={
            black,
            decoration={
                markings,
                mark = between positions 0 and 1 step 4mm with {
                    \\node[rectangle,fill=black,inner sep=0mm,minimum size=1mm]{};
                },
            },
            postaction={decorate}
        }
    }
    \\tikzset{decothree/.style={
            green,
            decoration={
                markings,
                mark = between positions 0 and 1 step 4mm with {
                    \\node[isosceles triangle,fill=green,inner sep=0mm,minimum size=1.5mm]{};
                },
            },
            postaction={decorate}
        }
    }
    \\tikzset{decofour/.style={
            red,
            decoration={
                markings,
                mark = between positions 0 and 1 step 4mm with {
                    \\node[diamond,fill=red,inner sep=0mm,minimum size=1.5mm]{};
                },
            },
            postaction={decorate}
        }
    }
    \\tikzset{decofive/.style={
            blue,
            decoration={
                markings,
                mark = between positions 0 and 1 step 4mm with {
                    \\node[circle,fill=blue,inner sep=0mm,minimum size=1mm]{};
                },
            },
            postaction={decorate}
        }
    }
"""

STYLES = [
    "decoone",
    "decotwo",
    "decothree",
    "decofour",
    "decofive"
]

SUBFIG_WIDTH = "0.48"
SUBFIG_SCALE= "0.7"

@dataclass
class FileData:
    total_num : int = 0
    num_compat : int = 0
    num_not_easy : int = 0
    num_unsat : int = 0
    num_approx_sat : int = 0
    num_unsolved_us : int = 0
    num_error : int = 0

def get_solvers(df : pd.DataFrame) -> List[str]:
    return [
        header[:-len(RESULTS_SUFFIX)]
        for header in df if header.endswith(RESULTS_SUFFIX)
    ]

def make_summary_data(df : pd.DataFrame):
    summary_data = FileData()

    df_compat = df[df[SYMPARIKH + RESULTS_SUFFIX] != INCOMPATIBLE]
    df_hard = df_compat[df_compat[Z3 + TIME_SUFFIX] != EASY_SAT]
    df_hard = df_hard[df_hard[Z3 + TIME_SUFFIX] != EASY_UNSAT]

    solvers = get_solvers(df_hard)

    df_unsat = df_hard[pd.concat({
        solver : df_hard[solver + RESULTS_SUFFIX] == UNSAT
        for solver in solvers
    }, axis="columns").any(axis="columns")]

    summary_data.total_num = len(df)
    summary_data.num_compat = len(df_compat)
    summary_data.num_not_easy = len(df_hard)
    summary_data.num_unsat = len(df_unsat)
    summary_data.num_approx_sat = len(df_unsat[
        df_unsat[SYMPARIKH + RESULTS_SUFFIX] == SAT
    ])
    summary_data.num_unsolved_us = len(df_hard[
        df_hard[SYMPARIKH + RESULTS_SUFFIX] == NO_RESULT
    ])
    summary_data.num_error = len(df_compat[
        df_compat[SYMPARIKH + RESULTS_SUFFIX] == ERROR
    ])

    df_sat = df_hard[pd.concat({
        solver : df_hard[solver + RESULTS_SUFFIX] == SAT
        for solver in solvers
        if solver != "sym_parikh"
    }, axis="columns").any(axis="columns")]
    df_wrong = df_sat[df_sat["sym_parikh_result"] == UNSAT]
    if len(df_wrong) > 0:
        print(f"Possible wrong answers in {filename}")
        for benchmark in df_wrong["filename"]:
            print(benchmark)

    return summary_data

def is_float(val : str) -> bool:
    try:
        float(val)
        return True
    except:
        return False

def make_solver_time_data(df : pd.DataFrame, solver : str) -> pd.DataFrame:
    time_col = solver + TIME_SUFFIX
    time_series = df[df[time_col].apply(is_float)][time_col].astype(float)
    time_series = time_series[time_series >= 0]
    time_series = time_series.sort_values(ignore_index=True)
    time_series = time_series.reset_index(drop=True)
    time_df = pd.concat({
        TIME: time_series,
        COUNT: time_series.index.to_series().astype(int).add(1)
    }, axis="columns")
    time_df = time_df.drop_duplicates(subset=[TIME], keep="last")
    return time_df

def make_time_data(df : pd.DataFrame) -> pd.DataFrame:
    solvers = get_solvers(df)
    all_solvers = pd.concat({
        solver : make_solver_time_data(df, solver)
        for solver in solvers
    }, axis="columns")
    all_solvers.columns = [
        '_'.join(col).strip() for col in all_solvers.columns.values
    ]
    return all_solvers

styles = dict()
def get_style(solver : str) -> str:
    if solver not in styles:
        style = STYLES.pop()
        styles[solver] = style
    return styles[solver]

def get_latex_graph(df_file : str, solvers : List[str]) -> str:
    return (
        "\\pgfplotstableread{" + time_data_file + "}{\\table}\n" +
        "\\begin{tikzpicture}[scale=" + SUBFIG_SCALE + "]\n" +
        "\\begin{axis}[\n" +
        "    xlabel={Time},\n" +
        "    ylabel={Solved Instances},\n" +
        "    legend pos=north west,\n" +
        "    grid=both\n" +
        "]\n" +
        "".join(
            "\\addplot[" + get_style(solver) + "] table " +
                "[x = {" + solver + TIME_SUFFIX +"}, " +
                 "y = {" + solver + COUNT_SUFFIX + "}] {\\table};\n"
            for solver in solvers
        ) +
        "\\end{axis}\n" +
        "\\end{tikzpicture}\n"
    )

def get_latex_summary(summary_datas : Dict[str, FileData]) -> str:
    latex = (
        "\\begin{table}\n"
        "\\begin{tabular}{lccccccc}\n"
            "\\toprule\n"
            "Set & "
            "Instances & "
            "Compatible & "
            "Hard & "
            "Unsat & "
            "Approx Sat & "
            "Unsolved & "
            "Error \\\\\n"
            "\\midrule\n"
    )

    for bset, data in summary_datas.items():
        latex += " & ".join(map(str, [
            bset, data.total_num, data.num_compat, data.num_not_easy,
            data.num_unsat, data.num_approx_sat, data.num_unsolved_us,
            data.num_error
        ])) + "\\\\\n"

    latex += (
        "\\bottomrule\n"
        "\\end{tabular}\n"
        "\\caption{Summary data for benchmark sets.}\n"
        "\\end{table}\n"
    )

    return latex

def escape_tex(text : str) -> str:
    return text.replace("_", "\\_")

def get_full_latex(
    graphs : Dict[str, str],
    summary_datas : Dict[str, FileData],
    solvers : Set[str]
) -> str:
    latex = (
        "\\documentclass{article}\n" +
        "\\usepackage{pgfplots}\n" +
        "\\usepackage{subcaption}\n" +
        "\\usepackage{booktabs}\n" +
        "\\usepackage{tikz}\n" +
        "\\usetikzlibrary{shapes,automata,positioning,decorations.markings}" +
        "\\begin{document}\n" +
        "\\begin{figure}\n" +
        STYLES_DEFINITIONS + "\n"
    )

    bsets = list(sorted(graphs))
    for bset in bsets:
        latex += (
            "\\begin{subfigure}{" + SUBFIG_WIDTH + "\\linewidth}\n" +
            graphs[bset] +
            "\\caption{" + bset + "}\n" +
            "\\end{subfigure}\n"
        )

    sorted_solvers = list(sorted(solvers))

    # add key
    latex += (
        "\\begin{center}\n" +
        "\\begin{tikzpicture}\n" +
        "\\begin{axis}[\n" +
        "   hide axis,\n" +
        "   xmin=50,\n" +
        "   xmax=50,\n" +
        "   ymin=0,\n" +
        "   ymax=0,\n" +
        "   legend columns=" + str(len(solvers)) + ",\n" +
        "   legend style={" +
                "/tikz/every even column" +
                "/.append style={column sep=0.5cm}}\n" +
        "]\n" +
        "".join(
            "\\addplot[" + get_style(solver) + "] {x};\n"
            for solver in sorted_solvers
        ) +
        "\\legend{" + ",".join(map(escape_tex, sorted_solvers)) + "}\n"
        "\\end{axis}\n" +
        "\\end{tikzpicture}\n" +
        "\\end{center}\n"
    )

    latex += (
        "\\end{figure}\n" +
        get_latex_summary(summary_datas) +
        "\\end{document}\n"
    )
    return latex

os.makedirs(RESULTS_DIR, exist_ok=True)

filedatas = dict()
graphs = dict()
solvers = set()

for filename in sys.argv[1:]:
    set_name = os.path.splitext(os.path.split(filename)[1])[0]

    df = pd.read_csv(filename)
    filedatas[filename] = make_summary_data(df)
    time_data = make_time_data(df)

    time_data_file = set_name + "_times.dat"
    time_data.to_csv(
        RESULTS_DIR + time_data_file,
        sep="\t",
        na_rep="nan",
        index_label="idx"
    )

    graphs[filename] = get_latex_graph(time_data_file, get_solvers(df))

    solvers.update(get_solvers(df))


with open(RESULTS_DIR + "graphs.tex", "w") as f:
    f.write(get_full_latex(graphs, filedatas, solvers))
