
import re

def _esc_chr(chrnum : int) -> str:
    return f"(_ char #x{str(hex(chrnum))[2:]})"

_SMTLIB_CATEGORY_WORD = "(re.union " + \
        "(re.range \"a\" \"z\") " + \
        "(re.range \"A\" \"Z\") " + \
        "(str.to_re \"_\")" + \
    ")"

_SMTLIB_CATEGORY_NOT_WORD = f"(re.comp {_SMTLIB_CATEGORY_WORD})"

_SMTLIB_CATEGORY_DIGIT = "(re.range \"0\" \"9\")"

_SMTLIB_CATEGORY_NOT_DIGIT = f"(re.comp {_SMTLIB_CATEGORY_DIGIT})"

_TAB = _esc_chr(ord('\t'))

_SMTLIB_CATEGORY_SPACE = "(re.union " + \
        "(str.to_re \" \") " + \
        f"(str.to_re {_TAB}) " + \
    ")"

_SMTLIB_CATEGORY_NOT_SPACE = f"(re.comp {_SMTLIB_CATEGORY_SPACE})"

def _parsed_regex_to_smtlib(regex) -> str:
    """Regex is from re._parser.parse"""
    if type(regex) == tuple:
        data = regex
    elif type(regex) == list:
        data = regex
    elif type(regex) == re._parser.SubPattern:
        data = regex.data
    else:
        raise ValueError(f"Unrecognised regex {regex}")

    if type(data) == list:
        parsed_data = list(filter(
            lambda res: res is not None and len(res.strip()) > 0,
            (_parsed_regex_to_smtlib(sub) for sub in data)
        ))
        if len(parsed_data) == 0:
            return ""
        elif len(parsed_data) == 1:
            return parsed_data[0]
        else:
            return (
                "(re.++ " +
                  " ".join(parsed_data) +
                ")"
            )
    elif type(data) == tuple:
        tag = data[0]
        if tag == re._constants.LITERAL:
            return f"(str.to_re {_esc_chr(data[1])})"
        if tag == re._constants.NOT_LITERAL:
            return f"(re.comp (str.to_re \"{_esc_chr(data[1])}\"))"
        if tag == re._constants.ANY:
            return "re.allchar"
        elif tag == re._constants.IN:
            alts = data[1]
            if len(alts) == 1:
                return _parsed_regex_to_smtlib(alts[0])
            else:
                if alts[0][0] == re._constants.NEGATE:
                    return (
                        "(re.comp (re.union " +
                            " ".join(
                                _parsed_regex_to_smtlib(sub)
                                for sub in alts[1:]
                            ) +
                        "))"
                    )

                else:
                    return (
                        "(re.union " +
                            " ".join(
                                _parsed_regex_to_smtlib(sub)
                                for sub in alts
                            ) +
                        ")"
                    )
        elif tag == re._constants.RANGE:
            lohi = data[1]
            lo = _esc_chr(lohi[0])
            hi = _esc_chr(lohi[1])
            return f"(re.range {lo} {hi})"
        elif tag in [
            re._constants.MAX_REPEAT,
            re._constants.MIN_REPEAT,
            re._constants.POSSESSIVE_REPEAT,
        ]:
            # approximate MIN_REPEAT as MAX_REPEAT for experiments
            tag_data = data[1]
            lo = tag_data[0]
            hi = tag_data[1]
            subre = tag_data[2]
            if hi == re._constants.MAXREPEAT:
                if lo == 0:
                    return f"(re.* {_parsed_regex_to_smtlib(subre)})"
                elif lo > 0:
                    # approximate larger lo numbers with re.+
                    return f"(re.+ {_parsed_regex_to_smtlib(subre)})"
                else:
                    raise ValueError(f"Unrecognised lo value {lo}")
            elif type(lo) == int and type(hi) == int:
                return f"((_ re.loop {lo} {hi}) {_parsed_regex_to_smtlib(subre)})"
        elif tag == re._constants.BRANCH:
            tag_data = data[1]
            if tag_data[0] is not None:
                raise ValueError(f"Unrecognised branch tag {tag_data}")
            else:
                return (
                    "(re.union " +
                        " ".join(
                            _parsed_regex_to_smtlib(sub)
                            for sub in tag_data[1]
                        ) +
                    ")"
                )
        elif tag == re._constants.AT:
            tag_data = data[1]
            # don't support anchors
            if tag_data in [
                re._constants.AT_END,
                re._constants.AT_BEGINNING,
                re._constants.AT_BOUNDARY,
                re._constants.AT_NON_BOUNDARY,
                re._constants.AT_BEGINNING_STRING,
                re._constants.AT_END_STRING,
            ]:
                return ""
            else:
                raise ValueError(f"Unrecognised AT {tag_data}")
        elif tag == re._constants.SUBPATTERN:
            tag_data = data[1]
            # Ignore extra info:
            #  tag_data[0] is capture group number (backref not supported)
            #  tag_data[1] is flags, i think. E.g. (?i: ...) is 2
            #  tag_data[2] i don't know
            return _parsed_regex_to_smtlib(tag_data[3])
        elif tag == re._constants.CATEGORY:
            tag_data = data[1]
            if tag_data == re._constants.CATEGORY_WORD:
                return _SMTLIB_CATEGORY_WORD
            if tag_data == re._constants.CATEGORY_NOT_WORD:
                return _SMTLIB_CATEGORY_NOT_WORD
            if tag_data == re._constants.CATEGORY_DIGIT:
                return _SMTLIB_CATEGORY_DIGIT
            if tag_data == re._constants.CATEGORY_NOT_DIGIT:
                return _SMTLIB_CATEGORY_NOT_DIGIT
            if tag_data == re._constants.CATEGORY_SPACE:
                return _SMTLIB_CATEGORY_SPACE
            if tag_data == re._constants.CATEGORY_NOT_SPACE:
                return _SMTLIB_CATEGORY_NOT_SPACE
            else:
                raise ValueError(f"Unrecognised category {tag_data}")
        elif tag == re._constants.ASSERT:
            # don't support lookahead
            return ""
        elif tag == re._constants.ASSERT_NOT:
            # don't support lookahead
            return ""
        elif tag == re._constants.GROUPREF:
            # approximate backreferences with .*
            return "(re.* re.anychar)"
        elif tag == re._constants.ATOMIC_GROUP:
            return _parsed_regex_to_smtlib(data[1])
        else:
            raise ValueError(f"Unrecognised regex tag {tag} in {regex}")
    else:
        raise ValueError(f"Unrecognised regex data {data}")

def regex_to_smtlib(regex : str) -> str:
    return _parsed_regex_to_smtlib(re._parser.parse(regex))

