
import csv
import glob
import re
import subprocess
import sys
import time

from dataclasses import dataclass
from typing import List, Optional

if len(sys.argv) < 2:
    print("Usage: ./test-solvers.py <dir containing benchmarks>")
    exit(-1)

benchmark_directory = sys.argv[1]

SYM_PARIKH = "/home/matt/research/strings/symbolic-parikh/symparikh-release/build/sym_parikh"

@dataclass
class Solver:
    command : str
    args : List[str]

Z3 = Solver("z3", [])
SOLVERS = {
    "z3": Z3,
    "ostrich": Solver("/home/matt/research/strings/ostrich/ostrich-client", []),
    "ostrich-cea": Solver(
        "/home/matt/research/strings/ostrich/ostrich-cea",
        ["+parikh", "-portfolio=strings"]
    ),
    "cvc5": Solver("cvc5", []),
    "sym_parikh": Solver(SYM_PARIKH, ["--file"])
}

# How long Z3 has to take for an example to be considered hard
HARD_TIMEOUT = 10 # seconds
# How long each solver gets to do a run
RUN_TIMEOUT = 30 # seconds
# It's a problem if the compat check takes more than this
COMPAT_TIMEOUT = 30

@dataclass
class SolverResult:
    check_result : str
    time : float
    timeout : bool

def check_compat(filename : str) -> Optional[bool]:
    try:
        res = subprocess.run(
            [SYM_PARIKH, "--check-compat", "--file", filename],
            capture_output=True,
            timeout=COMPAT_TIMEOUT
        )
        if res.returncode != 0:
            return None
        else:
            return "File compatible" in str(res.stdout)
    except subprocess.TimeoutExpired:
        return None

def run_solver(solver : Solver, filename : str, timeout : int) -> SolverResult:
    try:
        start_time = time.time()
        res = subprocess.run(
            [solver.command] + solver.args + [filename],
            timeout=timeout,
            capture_output=True
        )
        end_time = time.time()
        find_result = re.search(
            r"^(sat|unsat|unknown)$", res.stdout.decode("utf-8"), re.MULTILINE
        )
        check_result = "" if find_result is None else find_result[0]
        return SolverResult(check_result, end_time - start_time, False)
    except subprocess.TimeoutExpired:
        return SolverResult("", -1, True)

def main(benchmark_directory : str):
    writer = csv.writer(sys.stdout)
    writer.writerow(["filename", "compat_timeout"] + [
        field
        for solver in sorted(SOLVERS)
        for field in [f"{solver}_result", f"{solver}_time"]
    ])
    globs = glob.glob(f"{benchmark_directory}/**/*.smt2", recursive=True)
    for filename in globs:
        compat = check_compat(filename)
        if compat is not None and not compat:
            writer.writerow(
                [filename] + ["Incompatible"] * 2 * len(SOLVERS)
            )
            continue
        else:
            row = [filename]
            if compat is None:
                row.append("Timeout")
            else:
                row.append("OK")

            z3_res = run_solver(Z3, filename, HARD_TIMEOUT)
            if z3_res.timeout:
                for name in sorted(SOLVERS):
                    solver = SOLVERS[name]
                    res = run_solver(solver, filename, RUN_TIMEOUT)
                    row += [res.check_result, res.time]
                writer.writerow(row)
            else:
                writer.writerow(
                    row +
                    [f"Easy ({z3_res.check_result})"] * 2 * len(SOLVERS)
                )

main(benchmark_directory)
