
#include <chrono>
#include <iostream>
#include <stdlib.h>

#include "z3++.h"

#include "command_line.h"
#include "parikh_image_provider.h"
#include "sym_parikh.h"

using namespace std;
using namespace sym_parikh;
using namespace z3;

int main(int argc, char **argv) {
    CommandLine command_line(argc, argv);

    if (command_line.get_is_help()) {
        cerr << command_line.get_usage() << endl;
    } else {
        shared_ptr<string> const smt_filename = command_line.get_smt_filename();

        if (smt_filename != NULL) {
            try {
                auto start_time = chrono::high_resolution_clock::now();

                cout << "Building formula..." << endl;

                SymParikh sym_parikh = SymParikh::from_file(
                    *smt_filename,
                    SymParikhConfig(
                        command_line.get_pi_provider(),
                        command_line.get_predicates_filename(),
                        command_line.get_preds_from_lits(),
                        command_line.get_get_model(),
                        command_line.get_predicate_num_limit()
                    )
                );

                auto aut_time = chrono::high_resolution_clock::now();

                if (command_line.get_print_predicates()) {
                    for (expr const &e : sym_parikh.get_predicates())
                        cout << "Predicate: " << e << endl;
                    cout.flush();
                }

                if (command_line.get_check_compatibility()) {
                    cout << "File compatible" << endl;
                } else {
                    auto build_time_taken =
                        chrono::duration_cast<chrono::milliseconds>(
                            aut_time - start_time
                        ).count();

                    cout << "Checking formula..." << endl;
                    auto check_res = sym_parikh.check_parikh_sat();

                    auto end_time = chrono::high_resolution_clock::now();
                    auto time_taken =
                        chrono::duration_cast<chrono::milliseconds>(
                            end_time - start_time
                        ).count();

                    cout << check_res << "\n"
                        << "("
                            << "Total " << time_taken << "ms; "
                            << "Formula build " << build_time_taken << "ms"
                        << ")"
                        << endl;

                    if (check_res == sat && command_line.get_get_model()) {
                        shared_ptr<SymParikhModel> model =
                            sym_parikh.get_model();
                        if (model) {
                            for (expr const &s : model->get_string_vars()) {
                                cout << s << ":" << "\n";
                                for (expr const &p : model->get_predicates()) {
                                    cout << "  "
                                        << p << ": "
                                        << model->get_count(s, p) << "\n";
                                }
                            }
                            cout << endl;
                        } else {
                            cout << "No model." << endl;
                        }
                    }
                }
            } catch (SymParikhException *e) {
                cout << "File not compatible: " << e->what() << endl;
                return command_line.get_check_compatibility() ? 0 : -1;
            }
        } else {
            cout << "No file specified, try --help." << endl;
        }
    }

    return 0;
}
