
#include <cmath>
// for easy debugging
#include <iostream>

#include "z3++.h"

#include "parikh_image_one.h"
#include "re2automaton.h"
#include "sym_parikh_util.h"

using namespace std;
using namespace sym_parikh;
using namespace z3;

ParikhImageOne::ParikhImageOne(
    class context &context,
    shared_ptr<symbolic_automaton> const automaton,
    vector<expr> const &predicates,
    vector<expr> const &pred_vars
) : ParikhImage(context),
        re_automaton(automaton),
        move_vbles(),
        state_vbles(),
        thin_air_vbles(),
        final_vbles(),
        char_vbles(),
        char_count_vbles(),
        predicates(predicates),
        pred_vars(pred_vars),
        num_chars_by_tran(),
        image_formula(context.bool_val(true)),
        labels_to_moves() {
    assert(predicates.size() == pred_vars.size());

    expr move_image_formula = build_move_image_formula();

    build_labels_to_moves_map();

    expr correct_chars_formula = build_correct_chars_formula();
    expr correct_sums_formula = build_correct_char_counts_formula();
    expr correct_move_counts_formula = build_correct_move_counts_formula();
    expr var_domains = build_var_domains_formula();

    image_formula = exists(
        get_internal_variables(),
        move_image_formula
        && correct_chars_formula
        && correct_sums_formula
        && correct_move_counts_formula
        && var_domains
    );
}

expr const ParikhImageOne::build_move_image_formula() {
    expr const in_flow = make_in_flow_equations();
    expr const out_flow = make_out_flow_equations();
    expr const thin_air = make_thin_air_equations();
    expr const finals = make_finals_equations();

    return in_flow && out_flow && thin_air && finals;
}

void ParikhImageOne::build_labels_to_moves_map() {
    for (unsigned s = 0; s <= re_automaton->max_state(); s++) {
        if (re_automaton->is_state(s)) {
            for (auto m : re_automaton->get_moves_from(s)) {
                if (m.t().has_value()) {
                    unsigned label_id = m.t().value().id();
                    if (labels_to_moves.find(label_id) == labels_to_moves.end()) {
                        vector<symbolic_automaton::move> moves;
                        labels_to_moves.insert_or_assign(label_id, moves);
                    }
                    labels_to_moves.find(label_id)->second.push_back(m);
                }
            }
        }
    }
}

expr const ParikhImageOne::build_correct_chars_formula() {
    expr chars_fmla = context.bool_val(true);
    expr fv = get_free_variable_var();

    for (auto const &entry : labels_to_moves) {
        vector<symbolic_automaton::move> const &moves = entry.second;
        if (moves.size() > 0 && moves[0].t().has_value()) {
            expr const label = moves[0].t().value();
            unsigned num_chars = get_num_char_vectors(label);
            for (int cnum = 0; cnum < num_chars; cnum++) {
                expr char_vble = getadd_label_char(label, cnum);
                expr good = sub(context, label, fv, char_vble);
                chars_fmla = chars_fmla && good;
            }
        }
    }

    return chars_fmla;
}

expr const ParikhImageOne::build_correct_char_counts_formula() {
    // sum of chars matches predicate counts
    expr const fv = get_free_variable_var();
    expr const zero = context.int_val(0);

    vector<expr> pred_sums;
    for (int p = 0; p < pred_vars.size(); p++)
        pred_sums.push_back(context.int_val(0));

    for (auto const &entry : labels_to_moves) {
        vector<symbolic_automaton::move> const &moves = entry.second;
        if (moves.size() > 0 && moves[0].t().has_value()) {
            expr label = moves[0].t().value();
            unsigned num_chars = get_num_char_vectors(label);

            for (unsigned p = 0; p < pred_vars.size(); p++) {
                for (unsigned cnum = 0; cnum < num_chars; cnum++) {
                    expr char_vble = getadd_label_char(label, cnum);
                    expr char_count_vble = getadd_label_char_count(label, cnum);
                    expr good = sub(context, predicates[p], fv, char_vble);
                    expr contrib = ite(good, char_count_vble, zero);
                    pred_sums[p] = pred_sums[p] + contrib;
                }
            }
        }
    }

    expr counts_fmla = context.bool_val(true);
    for (unsigned p = 0; p < pred_vars.size(); p++)
        counts_fmla = counts_fmla && pred_vars[p] == pred_sums[p];

    return counts_fmla;
}

expr const ParikhImageOne::build_correct_move_counts_formula() {
    expr counts_fmla = context.bool_val(true);

    // Need 0 test else we end up asserting move used 0 times
    if (predicates.size() == 0)
        return counts_fmla;

    // sum of label char counts = sum of transitions with label
    for (auto const &entry : labels_to_moves) {
        vector<symbolic_automaton::move> const &moves = entry.second;
        if (moves.size() > 0 && moves[0].t().has_value()) {
            expr label = moves[0].t().value();
            unsigned num_chars = get_num_char_vectors(label);

            expr char_sum = context.int_val(0);
            for (int cnum = 0; cnum < num_chars; cnum++) {
                char_sum = char_sum + getadd_label_char_count(label, cnum);
            }
            expr moves_sum = context.int_val(0);
            for (auto const &m : moves) {
                moves_sum = moves_sum + getadd_move_count_vble(m);
            }

            counts_fmla = counts_fmla & (char_sum == moves_sum);
        }
    }

    return counts_fmla;
}

expr const ParikhImageOne::build_var_domains_formula() {
    expr fv = get_free_variable_var();
    expr domains = context.bool_val(true);

    // counts >= 0
    for (unsigned s = 0; s <= re_automaton->max_state(); s++) {
        if (re_automaton->is_state(s)) {
            expr state_count = getadd_state_count_vble(s);
            domains = domains && (state_count >= 0);
            for (auto m : re_automaton->get_moves_from(s)) {
                expr move_count = getadd_move_count_vble(m);
                domains = domains && (move_count >= 0);
            }
        }
    }

    // each set of move labels are such that
    //
    // * diff(l1, l2) or count2 == 0 (don't repeat same label)
    // * diff(l1, l2) means l1 < l2 and both differ on at least one
    // predicate
    // * counti == 0 => counti+1 == 0
    for (auto const &entry : labels_to_moves) {
        vector<symbolic_automaton::move> const &moves = entry.second;
        if (moves.size() > 0 && moves[0].t().has_value()) {
            expr label = moves[0].t().value();
            unsigned num_chars = get_num_char_vectors(label);

            for (int cnum = 0; cnum < num_chars; cnum++) {
                expr ccount = getadd_label_char_count(label, cnum);
                domains = domains && (ccount >= 0);
                if (cnum > 0) {
                    expr prev_ccount = getadd_label_char_count(label, cnum - 1);
                    expr prev_char = getadd_label_char(label, cnum - 1);
                    expr curr_char = getadd_label_char(label, cnum);

                    // either different to prev or ccount is 0
                    domains = domains && (
                        (prev_char < curr_char)
                        || (ccount == 0 && prev_char == curr_char)
                    );
                    expr some_diff = context.bool_val(false);
                    for (expr const &pred : predicates) {
                        some_diff = some_diff ||
                            sub(context, pred, fv, prev_char) !=
                                sub(context, pred, fv, curr_char);
                    }
                    // some predicate diff or ccount is 0
                    domains = domains && (some_diff || ccount == 0);
                    // ccount is 0 if previous is
                    domains = domains && (prev_ccount > 0 || ccount == 0);
                }
            }
        }
    }

    return domains;
}

expr const ParikhImageOne::getadd_move_count_vble(
    symbolic_automaton::move const &move
) {
    auto value = move_vbles.find(move);
    if (value == move_vbles.end()) {
        string name = "t_"
            + to_string(move.src())
            + "_"
            + to_string(move.dst());
        if (move.t().has_value()) {
            name += "_e" + to_string(move.t().value().id());
        }
        expr vble = context.int_const(name.c_str());
        move_vbles.insert({move, vble});
        return vble;
    } else {
        return value->second;
    }
}

expr const ParikhImageOne::getadd_label_char(expr const &label, int cnum) {
    auto key = make_pair(label.id(), cnum);
    auto value = char_vbles.find(key);
    if (value == char_vbles.end()) {
        string name = "l_" + to_string(label.id()) + "_a_" + to_string(cnum);
        expr vble = make_char_var(name);
        char_vbles.insert({key, vble});
        return vble;
    } else {
        return value->second;
    }
}

expr const ParikhImageOne::getadd_label_char_count(expr const &label, int cnum) {
    auto key = make_pair(label.id(), cnum);
    auto value = char_count_vbles.find(key);
    if (value == char_count_vbles.end()) {
        string name = "l_" + to_string(label.id()) + "_c_" + to_string(cnum);
        expr vble = context.int_const(name.c_str());
        char_count_vbles.insert({key, vble});
        return vble;
    } else {
        return value->second;
    }
}

expr const ParikhImageOne::getadd_kind_of_state_vble(
    map<unsigned, expr> &variables,
    unsigned state,
    string const &prefix
) {
    auto value = variables.find(state);
    if (value == variables.end()) {
        string name = prefix + to_string(state);
        expr vble = context.int_const(name.c_str());
        variables.insert({state, vble});
        return vble;
    } else {
        return value->second;
    }
}

expr const ParikhImageOne::getadd_state_count_vble(unsigned state) {
    return getadd_kind_of_state_vble(state_vbles, state, "s_");
}

expr const ParikhImageOne::getadd_thin_air_vble(unsigned state) {
    return getadd_kind_of_state_vble(thin_air_vbles, state, "z_");
}

expr const ParikhImageOne::getadd_final_vble(unsigned state) {
    return getadd_kind_of_state_vble(final_vbles, state, "f_");
}

expr const ParikhImageOne::make_in_flow_equations() {
    expr flows = context.bool_val(true);
    for (unsigned s = 0; s <= re_automaton->max_state(); s++) {
        if (re_automaton->is_state(s)) {
            expr svar = getadd_state_count_vble(s);
            expr sum = context.int_val(re_automaton->init() == s ? 1 : 0);
            for (auto m : re_automaton->get_moves_to(s)) {
                sum = sum + getadd_move_count_vble(m);
            }
            flows = flows && sum == svar;
        }
    }
    return flows;
}

expr const ParikhImageOne::make_out_flow_equations() {
    expr flows = context.bool_val(true);
    for (unsigned s = 0; s <= re_automaton->max_state(); s++) {
        if (re_automaton->is_state(s)) {
            expr svar = getadd_state_count_vble(s);
            expr sum = context.int_val(0);
            for (auto m : re_automaton->get_moves_from(s)) {
                sum = sum + getadd_move_count_vble(m);
            }
            // one less outgoing that incoming if final
            if (re_automaton->is_final_state(s))
                sum = sum + getadd_final_vble(s);
            flows = flows && sum == svar;
        }
    }
    return flows;
}

expr const ParikhImageOne::make_thin_air_equations() {
    expr thin_air = context.bool_val(true);

    for (unsigned s = 0; s <= re_automaton->max_state(); s++) {
        expr svar = getadd_state_count_vble(s);
        expr zvar = getadd_thin_air_vble(s);

        if (s == re_automaton->init()) {
            thin_air = thin_air && (zvar == 1);
        } else if (re_automaton->is_state(s)) {
            expr zero_or_reached = svar == 0 || zvar > 0;
            expr not_reached_and_zero = zvar == 0 && svar == 0;

            expr is_supported = context.bool_val(false);
            for (auto m : re_automaton->get_moves_to(s)) {
                expr tvar = getadd_move_count_vble(m);
                expr zsrc = getadd_thin_air_vble(m.src());

                is_supported = is_supported || (
                    zvar == zsrc + 1 && tvar > 0 && zsrc >= 1
                );
            }

            thin_air = thin_air && (
                zero_or_reached && (not_reached_and_zero || is_supported)
            );
        }
    }

    return thin_air;
}

expr const ParikhImageOne::make_finals_equations() {
    expr sum = context.int_val(0);
    for (unsigned s : re_automaton->final_states())
        sum = sum + getadd_final_vble(s);
    return sum == 1;
}

expr_vector ParikhImageOne::get_internal_variables() {
    expr_vector vars(context);
    get_variables_from_map(move_vbles, vars);
    get_variables_from_map(state_vbles, vars);
    get_variables_from_map(thin_air_vbles, vars);
    get_variables_from_map(final_vbles, vars);
    get_variables_from_map(char_vbles, vars);
    get_variables_from_map(char_count_vbles, vars);
    return vars;
}

unsigned ParikhImageOne::get_num_char_vectors(expr const &move_pred) {
    auto value = num_chars_by_tran.find(move_pred.id());
    if (value != num_chars_by_tran.end())
        return value->second;

    // upper bound
    unsigned dimension = predicates.size();
    unsigned ub_num_chars;
    if (dimension <= 1) ub_num_chars = 1;
    else if (dimension == 2) ub_num_chars = 2;
    else ub_num_chars = ceil(2 * dimension * log2(dimension));

    // now try to count

    solver solver(context);
    // move_pred always has to be true
    solver.add(move_pred);

    // merge two predicates if they can't be satisfied at same time
    // create merged list (p1, p2) (p3, p4, p5) (p6)
    // gives num of 3 * 4 * 2 chars.
    // I.e. sat p1 or p2 or neither = 3
    // sat p3 or p4 or p5 or neither = 4
    // ...

    expr tt = context.bool_val(true);

    vector<vector<expr>> blocks;
    for (expr const &pred : predicates) {
        // ignore true predicate as it doesn't create extra characters
        // (always true!)
        if (pred.id() == tt.id())
            continue;

        // ignore predicates that cannot overlap move_pred
        solver.push();
        solver.add(pred);
        check_result can_overlap = solver.check();
        solver.pop();
        if (can_overlap == unsat)
            continue;

        bool added = false;
        for (vector<expr> &block : blocks) {
            bool nooverlap = true;
            for (expr const &pred2 : block) {
                solver.push();
                solver.add(pred);
                solver.add(pred2);
                check_result res = solver.check();
                solver.pop();
                if (res == sat) {
                    nooverlap = false;
                    break;
                }
            }
            if (nooverlap) {
                block.push_back(pred);
                added = true;
                break;
            }
        }
        if (!added) {
            vector<expr> block;
            block.push_back(pred);
            blocks.push_back(block);
        }
    }

    unsigned count = 1;
    for (vector<expr> const &block : blocks) {
        unsigned block_size = block.size();
        // +1 if it's possible to satisfy none of the predicates in the
        // block
        solver.push();
        for (expr const &pred : block)
            solver.add(not(pred));
        if (solver.check() == sat)
            block_size += 1;
        solver.pop();

        count *= block_size;
    }

    if (count > ub_num_chars)
        count = ub_num_chars;

    num_chars_by_tran.insert({ move_pred.id(), count });

    return count;
}

