
#ifndef PARIKH_IMAGE_ONE_H
#define PARIKH_IMAGE_ONE_H

#include <map>

#include "z3++.h"

#include "parikh_image.h"
#include "symbolic_automaton.h"

namespace sym_parikh {
    /**
     * Construct a QF_LIA representation of the given regular expression
     *
     * This is the variant with more variables but possibly shorter
     * formulas.
     */
    class ParikhImageOne : public ParikhImage {
        public:
            /**
             * Build QF_LIA formula
             *
             * @param context z3 context to build formula with, keeps a
             * reference
             * @param automaton automaton to image
             * @param predicates vector of predicates that should be
             * counted
             * @param pred_vars the variables that should equal the
             * number of each predicate
             */
            ParikhImageOne(
                z3::context &context,
                std::shared_ptr<symbolic_automaton> const automaton,
                std::vector<z3::expr> const &predicates,
                std::vector<z3::expr> const &pred_vars
            );

            virtual z3::expr const get_image_formula() const {
                return image_formula;
            }

            virtual std::shared_ptr<symbolic_automaton const>
            get_re_automaton() {
                return re_automaton;
            }

        private:
            std::shared_ptr<symbolic_automaton> re_automaton;
            z3::expr image_formula;
            std::map<symbolic_automaton::move, z3::expr> move_vbles;
            std::map<unsigned, z3::expr> state_vbles;
            std::map<unsigned, z3::expr> thin_air_vbles;
            std::map<unsigned, z3::expr> final_vbles;
            /**
             * Maps from (move label expr id, char num) to variable
             */
            std::map<std::pair<unsigned, unsigned>, z3::expr>
                char_vbles;
            std::map<std::pair<unsigned, unsigned>, z3::expr>
                char_count_vbles;
            std::vector<z3::expr> predicates;
            std::vector<z3::expr> pred_vars;
            /**
             * The max number of chars for a given transition predicate
             * id
             */
            std::map<unsigned, unsigned> num_chars_by_tran;
            /**
             * Map transition label (ids) to the set of transition that
             * have them
             */
            std::map<unsigned, std::vector<symbolic_automaton::move>>
                labels_to_moves;

            /**
             * Builds a formula that counts (using move_vbles field)
             * number of each move
             */
            z3::expr const build_move_image_formula();

            /**
             * Get the map from transition labels to the transitions
             * that have them
             */
            void build_labels_to_moves_map();

            /**
             * Builds a formula that says each guessed character matches
             * its transition
             */
            z3::expr const build_correct_chars_formula();

            /**
             * Assert that pred is satisfied pred_var number of times
             *
             * That is for each transition, and for each predicate p
             * with count x that x = sum(p(ai)) where ai are the
             * characters on each of the num_chars transitions
             */
            z3::expr const build_correct_char_counts_formula();

            /**
             * Assert that each move used the right number of times
             *
             * That the char counts in total per move match the
             * number of times a move is used.
             */
            z3::expr const build_correct_move_counts_formula();

            /**
             * Assert that variables have values in allowed range
             *
             * That is, counts are non-negative, and characters are
             * singleton strings.
             *
             * @param dimension the number of predicates
             * @param num_chars the number of chars needed per transition
             */
            z3::expr const build_var_domains_formula();

            /**
             * Get or create the variable counting number of move
             */
            z3::expr const getadd_move_count_vble(
                symbolic_automaton::move const &move
            );

            /**
             * Get or create the variable counting number of state
             */
            z3::expr const getadd_state_count_vble(unsigned state);

            /**
             * Get or create the variable counting number of state
             */
            z3::expr const getadd_thin_air_vble(unsigned state);

            /**
             * Get or create the variable saying if state is the final
             * in the run
             */
            z3::expr const getadd_final_vble(unsigned state);

            /**
             * Get the cnum-th character for a transition
             */
            z3::expr const getadd_label_char(z3::expr const &label, int cnum);

            /**
             * Get the count of the cnum-th character for a transition
             */
            z3::expr const getadd_label_char_count(z3::expr const &label, int cnum);

            /**
             * Make the equations about flows into aut states
             *
             * I.e. s = sum(ti) where ti.dst() == s (+1?)
             * Assumes re_automaton has been built
             */
            z3::expr const make_in_flow_equations();

            /**
             * Make the equations about flows out of aut states
             *
             * I.e. s = sum(ti) where ti.src() == s
             * Assumes re_automaton has been built
             */
            z3::expr const make_out_flow_equations();

            /**
             * Make the "thin air checks" of aut states
             *
             * I.e. s is only positive if it's reached from the initial
             * state
             * Assumes re_automaton has been built
             */
            z3::expr const make_thin_air_equations();

            /**
             * Make the finals of aut states
             *
             * I.e. is only one final state (f1 + ... + fn = 1)
             * Assumes re_automaton has been built
             */
            z3::expr const make_finals_equations();

            /**
             * Implements basic pattern for getting a state variable
             */
            z3::expr const getadd_kind_of_state_vble(
                std::map<unsigned, z3::expr> &variables,
                unsigned state,
                std::string const &prefix
            );

            /**
             * Gets a vector of all internal variables to the encoding
             *
             * So they can be existentially hidden
             */
            z3::expr_vector get_internal_variables();

            /**
             * Estimate how many diff chars are possible with predicates
             *
             * I.e. how many characters are there that result in
             * different satisfaction of preds. E.g. if your preds are
             * (= x 'a') and (= x 'b'), there are only really three
             * chars (0, 0), (1, 0), and (0, 1), since nothing satisfies
             * both.
             *
             * Specialised to a transition. I.e. always have to satisfy
             * move predicate.
             */
            unsigned get_num_char_vectors(z3::expr const &move_pred);
    };
}

#endif
