
#include <cmath>
// for easy debugging
#include <iostream>

#include "z3++.h"

#include "parikh_image_two.h"
#include "re2automaton.h"
#include "sym_parikh_util.h"

using namespace std;
using namespace sym_parikh;
using namespace z3;

static unsigned get_num_aut_states(shared_ptr<symbolic_automaton> const aut) {
    unsigned count = 0;
    for (unsigned s = 0; s <= aut->max_state(); s++) {
        if (aut->is_state(s))
            count += 1;
    }
    return count;
}

ParikhImageTwo::ParikhImageTwo(
    class context &context,
    shared_ptr<symbolic_automaton> automaton,
    vector<expr> const &predicates,
    vector<expr> const &pred_vars
) : ParikhImage(context),
        re_automaton(automaton),
        qin_vbles(),
        a_vbles(),
        qout_vbles(),
        count_vbles(),
        sin_vbles(),
        sout_vbles(),
        final_vbles(),
        thin_air_vbles(),
        limited_move_start(),
        image_formula(context.bool_val(true)) {
    assert(predicates.size() == pred_vars.size());

    unsigned dimension = predicates.size();
    unsigned n = dimension + 2 * get_num_aut_states(re_automaton);
    unsigned num_moves = re_automaton->num_moves();
    num_limited_positions = ceil(2 * dimension * log2(dimension));
    unsigned max_chars_with_states = ceil(2 * n * log2(n));
    unsigned max_chars_by_move = num_moves * num_limited_positions;
    unsigned num_chars = min(max_chars_with_states, max_chars_by_move);

    allocate_move_limits(num_chars);

    expr good_guessed_moves = build_good_guessed_moves_formula(num_chars);
    expr good_state_counts = build_good_state_counts_formula(num_chars);
    expr good_predicate_counts = build_good_predicate_counts_formula(
        predicates,
        pred_vars,
        num_chars
    );
    expr is_run = build_is_run_formula(num_chars);
    expr var_domains = build_var_domains_formula(num_chars);

    image_formula = exists(
        get_internal_variables(),
        good_guessed_moves
        && good_state_counts
        && good_predicate_counts
        && is_run
        && var_domains
    );
}

void ParikhImageTwo::allocate_move_limits(unsigned num_chars) {
    int pos = 0;
    for (unsigned s = 0; s <= re_automaton->max_state(); s++) {
        if (re_automaton->is_state(s)) {
            for (auto m : re_automaton->get_moves_from(s)) {
                limited_move_start.insert({m, pos});
                pos += num_limited_positions;
                if (pos >= num_chars)
                    return;
            }
        }
    }
}

expr const ParikhImageTwo::build_good_guessed_moves_formula(unsigned num_chars) {
    expr all_good = context.bool_val(true);

    for (unsigned i = 0; i < num_chars; i++) {
        all_good = all_good && build_good_guessed_move_formula(i);
    }

    return all_good;
}

expr const ParikhImageTwo::build_good_guessed_move_formula(unsigned pos) {
    expr fv = get_free_variable_var();
    expr epsilon = get_epsilon();
    unsigned max_state = re_automaton->max_state();
    expr no_state = context.int_val(max_state + 1);

    expr qin = getadd_qin_vble(pos);
    expr a = getadd_char_vble(pos);
    expr qout = getadd_qout_vble(pos);

    expr good = qin == no_state && a == epsilon && qout == no_state;

    for (unsigned s = 0; s <= re_automaton->max_state(); s++) {
        if (re_automaton->is_state(s)) {
            for (auto m : re_automaton->get_moves_from(s)) {
                if (move_good_for_position(m, pos)) {
                    expr a_good = m.t().has_value()
                        ? sub(context, m.t().value(), fv, a)
                        : a == epsilon;

                    expr src = context.int_val(m.src());
                    expr dst = context.int_val(m.dst());

                    good = good || (qin == src && a_good && qout == dst);
                }
            }
        }
    }

    return good;
}

expr const ParikhImageTwo::build_good_state_counts_formula(unsigned num_chars) {
    expr zero = context.int_val(0);

    expr good = context.bool_val(true);

    for (unsigned s = 0; s <= re_automaton->max_state(); s++) {
        if (re_automaton->is_state(s)) {
            expr sin = getadd_sin_vble(s);
            expr sout = getadd_sout_vble(s);

            expr sin_sum = zero;
            expr sout_sum = zero;

            expr sfmla = context.int_val(s);

            // this gets a bit copy-pasty...

            set<unsigned> positions_in;
            bool all = state_good_for_positions_in(s, positions_in);

            if (all) {
                for (int pos = 0; pos < num_chars; pos++) {
                    expr qin = getadd_qin_vble(pos);
                    expr count = getadd_count_vble(pos);
                    sin_sum = sin_sum + ite(qin == sfmla, count, zero);
                }
            } else {
                for (unsigned pos : positions_in) {
                    expr qin = getadd_qin_vble(pos);
                    expr count = getadd_count_vble(pos);
                    sin_sum = sin_sum + ite(qin == sfmla, count, zero);
                }
            }

            set<unsigned> positions_out;
            all = state_good_for_positions_out(s, positions_out);

            if (all) {
                for (int pos = 0; pos < num_chars; pos++) {
                    expr qout = getadd_qout_vble(pos);
                    expr count = getadd_count_vble(pos);
                    sout_sum = sout_sum + ite(qout == sfmla, count, zero);
                }
            } else {
                for (unsigned pos : positions_out) {
                    expr qout = getadd_qout_vble(pos);
                    expr count = getadd_count_vble(pos);
                    sout_sum = sout_sum + ite(qout == sfmla, count, zero);
                }
            }

            good = good && (sin == sin_sum && sout == sout_sum);
        }
    }

    return good;
}

expr const ParikhImageTwo::build_good_predicate_counts_formula(
    vector<expr> const &predicates,
    vector<expr> const &pred_vars,
    unsigned num_chars
) {
    expr fv = get_free_variable_var();
    expr epsilon = get_epsilon();

    expr counts_fmla = context.bool_val(true);

    // that each predicate has the right number of counts
    for (int pred_num = 0; pred_num < predicates.size(); pred_num++) {
        expr pred_count = pred_vars[pred_num];
        expr sum_fmla = context.int_val(0);
        for (int pos = 0; pos < num_chars; pos++) {
            expr char_vble = getadd_char_vble(pos);
            expr char_count_vble = getadd_count_vble(pos);
            expr good = (
                char_vble != epsilon &&
                sub(context, predicates[pred_num], fv, char_vble)
            );
            expr contrib = ite(good, char_count_vble, context.int_val(0));
            sum_fmla = sum_fmla + contrib;
        }
        counts_fmla = counts_fmla && (pred_count == sum_fmla);
    }

    return counts_fmla;
}

expr const ParikhImageTwo::build_is_run_formula(unsigned num_chars) {
    return build_flow_formula() && build_thin_air_formula(num_chars);
}

expr const ParikhImageTwo::build_flow_formula() {
    expr good = context.bool_val(true);

    // in = out modulo init/final
    for (unsigned s = 0; s <= re_automaton->max_state(); s++) {
        if (re_automaton->is_state(s)) {
            expr sin = getadd_sin_vble(s);
            expr sout = getadd_sout_vble(s);

            expr in_bias = context.int_val(0);
            expr out_bias = context.int_val(0);

            if (s == re_automaton->init())
                in_bias = context.int_val(1);
            if (re_automaton->is_final_state(s))
                out_bias = getadd_final_vble(s);

            good = good && (sin + out_bias == sout + in_bias);
        }
    }

    // and only one final
    expr finals_sum = context.int_val(0);
    for (unsigned s = 0; s <= re_automaton->max_state(); s++) {
        if (re_automaton->is_state(s) && re_automaton->is_final_state(s)) {
            finals_sum = finals_sum + getadd_final_vble(s);
        }
    }

    return good && (finals_sum == context.int_val(1));
}

expr const ParikhImageTwo::build_thin_air_formula(unsigned num_chars) {
    expr thin_air = context.bool_val(true);

    for (unsigned s = 0; s <= re_automaton->max_state(); s++) {
        expr svar = getadd_sin_vble(s);
        expr zvar = getadd_thin_air_vble(s);

        if (s == re_automaton->init()) {
            thin_air = thin_air && (zvar == 1);
        } else if (re_automaton->is_state(s)) {
            expr zero_or_reached = svar == 0 || zvar > 0;
            expr not_reached_and_zero = zvar == 0 && svar == 0;

            expr is_supported = context.bool_val(false);

            // look for all incoming transitions and see if there is
            // some incoming state to support it
            // which means looking through all guessed (q, a, q')...
            for (auto m : re_automaton->get_moves_to(s)) {
                expr zsrc = getadd_thin_air_vble(m.src());

                unsigned pos_min = move_good_for_position_min(m);
                unsigned pos_max = move_good_for_position_max(m, num_chars);

                for (int pos = pos_min; pos < pos_max; pos++) {
                    expr count_var = getadd_count_vble(pos);
                    expr qin_var = getadd_qin_vble(pos);
                    expr qout_var = getadd_qout_vble(pos);

                    is_supported = is_supported || (
                        qin_var == context.int_val(m.src()) &&
                        qout_var == context.int_val(s) &&
                        zvar == zsrc + 1 &&
                        count_var > 0
                        && zsrc >= 1
                    );
                }
            }

            thin_air = thin_air && (
                zero_or_reached && (not_reached_and_zero || is_supported)
            );
        }
    }

    return thin_air;
}

expr const ParikhImageTwo::build_var_domains_formula(unsigned num_chars) {
    expr nonneg = context.bool_val(true);

    for (unsigned i = 0; i < num_chars; i++) {
        nonneg = nonneg &&
            getadd_qin_vble(i) >= 0 &&
            getadd_qout_vble(i) >= 0 &&
            getadd_count_vble(i) >= 0;
    }

    for (unsigned s = 0; s <= re_automaton->max_state(); s++) {
        if (re_automaton->is_state(s)) {
            nonneg = nonneg &&
                getadd_sin_vble(s) >= 0 &&
                getadd_sout_vble(s) >= 0 &&
                getadd_thin_air_vble(s) >= 0;
            if (re_automaton->is_final_state(s))
                nonneg = nonneg && getadd_final_vble(s) >= 0;
        }
    }

    return nonneg;
}

expr const ParikhImageTwo::getadd_int_vble(
    map<unsigned, expr> &vble_map,
    string name,
    unsigned index
) {
    auto value = vble_map.find(index);
    if (value == vble_map.end()) {
        string var_name = "" + name + "_" + to_string(index);
        expr vble = context.int_const(var_name.c_str());
        vble_map.insert({index, vble});
        return vble;
    } else {
        return value->second;
    }
}

expr const ParikhImageTwo::getadd_qin_vble(unsigned pos) {
    return getadd_int_vble(qin_vbles, "qin", pos);
}

expr const ParikhImageTwo::getadd_char_vble(unsigned pos) {
    auto value = a_vbles.find(pos);
    if (value == a_vbles.end()) {
        string var_name = "a_" + to_string(pos);
        expr vble = make_char_var(var_name);
        a_vbles.insert({pos, vble});
        return vble;
    } else {
        return value->second;
    }
}

expr const ParikhImageTwo::getadd_qout_vble(unsigned pos) {
    return getadd_int_vble(qout_vbles, "qout", pos);
}

expr const ParikhImageTwo::getadd_count_vble(unsigned pos) {
    return getadd_int_vble(count_vbles, "c", pos);
}

expr const ParikhImageTwo::getadd_sin_vble(unsigned state) {
    return getadd_int_vble(sin_vbles, "sin", state);
}

expr const ParikhImageTwo::getadd_sout_vble(unsigned state) {
    return getadd_int_vble(sout_vbles, "sout", state);
}

expr const ParikhImageTwo::getadd_final_vble(unsigned state) {
    return getadd_int_vble(final_vbles, "f", state);
}

expr const ParikhImageTwo::getadd_thin_air_vble(unsigned state) {
    return getadd_int_vble(thin_air_vbles, "z", state);
}

bool ParikhImageTwo::move_good_for_position(
    symbolic_automaton::move const &move,
    unsigned pos
) {
    auto start_limit = limited_move_start.find(move);
    if (start_limit == limited_move_start.end()) {
        return true;
    } else {
        unsigned start = start_limit->second;
        return start <= pos && pos < start + num_limited_positions;
    }
}

unsigned ParikhImageTwo::move_good_for_position_min(
    symbolic_automaton::move const &move
) {
    auto start_limit = limited_move_start.find(move);
    if (start_limit == limited_move_start.end())
        return 0;
    else
        return start_limit->second;
}

unsigned ParikhImageTwo::move_good_for_position_max(
    symbolic_automaton::move const &move,
    unsigned num_chars
) {
    auto start_limit = limited_move_start.find(move);
    if (start_limit == limited_move_start.end())
        return num_chars;
    else
        return start_limit->second + num_limited_positions;
}

bool ParikhImageTwo::state_good_for_positions_in(
    unsigned state,
    std::set<unsigned> &positions
) {
    for (auto m : re_automaton->get_moves_from(state)) {
        auto start_limit = limited_move_start.find(m);
        if (start_limit == limited_move_start.end())
            return true;

        unsigned start = start_limit->second;
        unsigned end = start + num_limited_positions;
        for (int pos = start; pos < end; pos++)
            positions.insert(pos);
    }
    return false;
}

bool ParikhImageTwo::state_good_for_positions_out(
    unsigned state,
    std::set<unsigned> &positions
) {
    for (auto m : re_automaton->get_moves_to(state)) {
        auto start_limit = limited_move_start.find(m);
        if (start_limit == limited_move_start.end())
            return true;

        unsigned start = start_limit->second;
        unsigned end = start + num_limited_positions;
        for (int pos = start; pos < end; pos++)
            positions.insert(pos);
    }
    return false;
}

expr_vector ParikhImageTwo::get_internal_variables() {
    expr_vector vars(context);
    get_variables_from_map(qin_vbles, vars);
    get_variables_from_map(a_vbles, vars);
    get_variables_from_map(qout_vbles, vars);
    get_variables_from_map(count_vbles, vars);
    get_variables_from_map(sin_vbles, vars);
    get_variables_from_map(sout_vbles, vars);
    get_variables_from_map(final_vbles, vars);
    get_variables_from_map(thin_air_vbles, vars);
    return vars;
}
