
#ifndef PARIKH_IMAGE_TWO_H
#define PARIKH_IMAGE_TWO_H

#include <limits>
#include <map>

#include "z3++.h"

#include "parikh_image.h"
#include "symbolic_automaton.h"

namespace sym_parikh {
    /**
     * Construct a QF_LIA representation of the given regular expression
     *
     * This is the variant with more variables but possibly shorter
     * formulas.
     */
    class ParikhImageTwo : public ParikhImage {
        public:
            /**
             * Build QF_LIA formula
             *
             * @param context z3 context to build formula with, keeps a
             * reference
             * @param automaton
             * @param predicates vector of predicates that should be
             * counted
             * @param pred_vars the variables that should equal the
             * number of each predicate
             */
            ParikhImageTwo(
                z3::context &context,
                std::shared_ptr<symbolic_automaton> const automaton,
                std::vector<z3::expr> const &predicates,
                std::vector<z3::expr> const &pred_vars
            );

            virtual z3::expr const get_image_formula() const {
                return image_formula;
            }

            virtual std::shared_ptr<symbolic_automaton const> get_re_automaton() {
                return re_automaton;
            }

        private:
            std::shared_ptr<symbolic_automaton> re_automaton;
            z3::expr image_formula;

            /**
             * The N (q, a, q') variables for the guessed augmented
             * letters, plus the count for the number of times it is
             * used
             */
            std::map<unsigned, z3::expr> qin_vbles;
            std::map<unsigned, z3::expr> a_vbles;
            std::map<unsigned, z3::expr> qout_vbles;
            std::map<unsigned, z3::expr> count_vbles;

            /**
             * The s^in and s^out counts for each state, plus the
             * variable to say if state is the last of the run, and the
             * "thin air" variables (z)
             */
            std::map<unsigned, z3::expr> sin_vbles;
            std::map<unsigned, z3::expr> sout_vbles;
            std::map<unsigned, z3::expr> final_vbles;
            std::map<unsigned, z3::expr> thin_air_vbles;

            /**
             * Limit which guessed chars moves can appear in
             *
             * If move appears in this map, with
             *  pos = limited_move_start[move]
             * it can only match (q, a, q') in positions pos through
             * pos + num_limited_positions (non-inclusive on upper end).
             * Else it can appear anywhere.
             */
            std::map<symbolic_automaton::move, unsigned> limited_move_start;
            unsigned num_limited_positions;

            /**
             * Formula says guessed moves exist and match automaton
             *
             * I.e. every (q, a, q') has a q--f-->q' with f(a) true
             */
            z3::expr const build_good_guessed_moves_formula(unsigned num_chars);

            /**
             * Formula says guessed move exists and matches automaton
             *
             * I.e. (q, a, q') has a q--f-->q' with f(a) true
             */
            z3::expr const build_good_guessed_move_formula(unsigned num_chars);

           /**
             * Formula says guessed moves tally with sin/sout
             */
            z3::expr const build_good_state_counts_formula(unsigned num_chars);

            z3::expr const build_good_predicate_counts_formula(
                std::vector<z3::expr> const &predicates,
                std::vector<z3::expr> const &pred_vars,
                unsigned num_chars
            );

            /**
             * Formula that says guessed chars make a run
             */
            z3::expr const build_is_run_formula(unsigned num_chars);

            /**
             * Formula says (essentially) sin = sout
             *
             * Except for init and final, which have a +1
             */
            z3::expr const build_flow_formula();

            /**
             * Formula that says we have no floating self-supporting
             * loops
             */
            z3::expr const build_thin_air_formula(unsigned num_chars);

            /**
             * Formula that says all vars in right domain
             *
             * E.g. counts >= 0 and chars are strings of length 1.
             */
            z3::expr const build_var_domains_formula(unsigned num_chars);

            /**
             * Gets the pos-th qin variable
             */
            z3::expr const getadd_qin_vble(unsigned pos);

            /**
             * Gets the pos-th char variable
             */
            z3::expr const getadd_char_vble(unsigned pos);

            /**
             * Gets the pos-th qout variable
             */
            z3::expr const getadd_qout_vble(unsigned pos);

            /**
             * Gets the counts for guessed moves
             */
            z3::expr const getadd_count_vble(unsigned pos);

            /**
             * Get or create the variable counting number in to state
             */
            z3::expr const getadd_sin_vble(unsigned state);

            /**
             * Get or create the variable counting number out of state
             */
            z3::expr const getadd_sout_vble(unsigned state);

            /**
             * Get or create the variable that's 1 if state is the final
             * state of the run, else 0
             */
            z3::expr const getadd_final_vble(unsigned state);

            /**
             * Get or create the variable that does thin air check
             */
            z3::expr const getadd_thin_air_vble(unsigned state);

            /**
             * Decide which transitions are limited to certain positions
             */
            void allocate_move_limits(unsigned num_chars);

            /**
             * True if move is allowed to be at guessed char position
             */
            bool move_good_for_position(
                symbolic_automaton::move const &move,
                unsigned pos
            );

            /**
             * First pos a move is good for (inclusive)
             */
            unsigned move_good_for_position_min(
                symbolic_automaton::move const &move
            );

            /**
             * Last pos a move is good for (exclusive)
             */
            unsigned move_good_for_position_max(
                symbolic_automaton::move const &move,
                unsigned num_chars
            );

            /**
             * Loads set with positions where state may be dst
             *
             * Or returns true if could be any position
             */
            bool state_good_for_positions_in(
                unsigned state,
                std::set<unsigned> &positions
            );

            /**
             * Loads set with positions where state may be src
             *
             * Or returns true if could be any position
             */
            bool state_good_for_positions_out(
                unsigned state,
                std::set<unsigned> &positions
            );

            z3::expr const getadd_int_vble(
                std::map<unsigned, z3::expr> &vble_map,
                std::string name,
                unsigned index
            );

            /**
             * Get all declared variables that should be hidden
             */
            z3::expr_vector get_internal_variables();

            z3::expr const get_epsilon() const {
                return context.int_val(
                    std::numeric_limits<int>::min()
                );
            }
    };
}

#endif
