
// for easy debugging
#include <iostream>
#include <set>
#include <sstream>

#include "parikh_profile.h"
#include "re2automaton.h"

static const char* decl_kind_to_string(const Z3_decl_kind& kind);

using namespace std;
using namespace sym_parikh;
using namespace z3;

void ParikhProfile::extract_data(expr const &formula) {
    if (seen_expr_ids.find(formula.id()) != seen_expr_ids.end())
        return;

    seen_expr_ids.insert(formula.id());

    if (formula.is_seq() && formula.is_const() && !formula.is_string_value()) {
        seq_vars.push_back(formula);
    } else if (formula.is_not() && formula.args().size() > 0) {
        switch (formula.args()[0].decl().decl_kind()) {
        case Z3_OP_SEQ_IN_RE:
            not_in_re_checks.push_back(formula);
            break;
        // can only support re versions of = op
        // consider sequence version
        //   (assert (= x "ab"))
        //   (assert (= y "ba"))
        //   (assert (not (= x y)))
        // this becomes unsat since x y have same Parikh Image, but
        // formula is clearly sat. Same holds for prefixof, suffixof,
        // etc.
        case Z3_OP_EQ:
            if (formula.args()[0].args().size() == 2) {
                if (
                    formula.args()[0].args()[0].is_seq() &&
                    formula.args()[0].args()[1].is_seq()
                ) {
                    ostringstream ss;
                    ss << "Unsupported operation "
                        << formula << " "
                        << decl_kind_to_string(formula.decl().decl_kind());
                    throw new ParikhProfileException(ss.str());
                } else if (
                    formula.args()[0].args()[0].is_re() &&
                    formula.args()[0].args()[1].is_re()
                ) {
                    re_notequalities.push_back(formula);
                }
            }
            break;
        // catch things we definitely don't support
        case Z3_OP_SEQ_REPLACE_RE:
        case Z3_OP_SEQ_REPLACE_RE_ALL:
        case Z3_OP_SEQ_REPLACE_ALL:
        case Z3_OP_SEQ_AT:
        case Z3_OP_SEQ_NTH:
        case Z3_OP_SEQ_INDEX:
        case Z3_OP_SEQ_LAST_INDEX:
        case Z3_OP_STR_TO_INT:
        case Z3_OP_INT_TO_STR:
        case Z3_OP_UBV_TO_STR:
        case Z3_OP_SBV_TO_STR:
        case Z3_OP_STR_TO_CODE:
        case Z3_OP_STR_FROM_CODE:
        case Z3_OP_STRING_LT:
        case Z3_OP_STRING_LE:
        case Z3_OP_SEQ_PREFIX:
        case Z3_OP_SEQ_SUFFIX:
        case Z3_OP_SEQ_CONTAINS: {
            ostringstream ss;
            ss << "Unsupported operation "
                << formula << " "
                << decl_kind_to_string(formula.decl().decl_kind());
            throw new ParikhProfileException(ss.str());
        }
        default:
            break;
        }
        for (expr const &arg : formula.args()[0].args()) {
            extract_data(arg);
        }
    } else {
        switch (formula.decl().decl_kind()) {
        case Z3_OP_SEQ_IN_RE:
            in_re_checks.push_back(formula);
            break;
        case Z3_OP_SEQ_LENGTH:
            has_length = true;
            seq_lens.push_back(formula);
            break;
        case Z3_OP_EQ:
            if (formula.args().size() == 2) {
                if (
                    formula.args()[0].is_seq() &&
                    formula.args()[1].is_seq()
                ) {
                    seq_equalities.push_back(formula);
                } else if (
                    formula.args()[0].is_re() &&
                    formula.args()[1].is_re()
                ) {
                    re_equalities.push_back(formula);
                }
            }
            break;
        case Z3_OP_SEQ_PREFIX:
        case Z3_OP_SEQ_SUFFIX:
        case Z3_OP_SEQ_CONTAINS:
            if (
                formula.args().size() == 2 &&
                formula.args()[0].is_seq() &&
                formula.args()[1].is_seq()
            ) {
                seq_equalities.push_back(formula);
            }
            break;
        // catch things we definitely don't support
        case Z3_OP_SEQ_REPLACE_RE:
        case Z3_OP_SEQ_REPLACE_RE_ALL:
        case Z3_OP_SEQ_REPLACE_ALL:
        case Z3_OP_SEQ_AT:
        case Z3_OP_SEQ_NTH:
        case Z3_OP_SEQ_INDEX:
        case Z3_OP_SEQ_LAST_INDEX:
        case Z3_OP_STR_TO_INT:
        case Z3_OP_INT_TO_STR:
        case Z3_OP_UBV_TO_STR:
        case Z3_OP_SBV_TO_STR:
        case Z3_OP_STR_TO_CODE:
        case Z3_OP_STR_FROM_CODE:
        case Z3_OP_STRING_LT:
        case Z3_OP_STRING_LE: {
            ostringstream ss;
            ss << "Unsupported operation "
                << formula << " "
                << decl_kind_to_string(formula.decl().decl_kind());
            throw new ParikhProfileException(ss.str());
        }
        default:
            break;
        }
        for (expr const &arg : formula.args()) {
            extract_data(arg);
        }
    }
}

void ParikhProfile::make_automata() {
    for (auto re_check : in_re_checks) {
        auto args = re_check.args();
        if (args.size() > 1) {
            auto re = args[1];
            if (positive_automata.find(re.id()) == positive_automata.end()) {
                shared_ptr<symbolic_automaton> aut = re2aut.translate(re);
                positive_automata.insert({re.id(), aut});
            }
        }
    }
    for (auto re_check : not_in_re_checks) {
        if (re_check.args().size() > 0) {
            auto in_re = re_check.args()[0];
            if (in_re.args().size() > 1) {
                auto re = in_re.args()[1];
                if (negative_automata.find(re.id()) == negative_automata.end()) {
                    shared_ptr<symbolic_automaton> aut =
                        re2aut.translate_negative(re);
                    negative_automata.insert({re.id(), aut});
                }
            }
        }
    }
    for (auto re_eq : re_equalities) {
        auto args = re_eq.args();
        for (expr const &re : re_eq.args()) {
            if (positive_automata.find(re.id()) == positive_automata.end()) {
                shared_ptr<symbolic_automaton> aut = re2aut.translate(re);
                positive_automata.insert({re.id(), aut});
            }
        }
    }
    for (auto re_noteq : re_notequalities) {
        if (re_noteq.args().size() > 0) {
            auto re_eq = re_noteq.args()[0];
            for (expr const &re : re_eq.args()) {
                // will need both positive and negative for translating
                // not equals (see sym_parikh.cpp)
                if (
                    positive_automata.find(re.id()) ==
                    positive_automata.end()
                ) {
                    shared_ptr<symbolic_automaton> aut = re2aut.translate(re);
                    positive_automata.insert({re.id(), aut});
                }

                if (
                    negative_automata.find(re.id()) ==
                    negative_automata.end()
                ) {
                    shared_ptr<symbolic_automaton> aut =
                        re2aut.translate_negative(re);
                    negative_automata.insert({re.id(), aut});
                }
            }
        }
    }
}

void ParikhProfile::find_predicates() {
    if (config.predicates_filename == nullptr) {
        set<unsigned> pred_ids;
        find_new_predicates_from_automata_map(positive_automata, pred_ids);
        find_new_predicates_from_automata_map(negative_automata, pred_ids);
        // get preds from lits if asked, or if we didn't find any in regexp
        if (config.preds_from_lits || pred_ids.size() == 0)
            find_new_predicates_from_lits(pred_ids);
    } else {
        expr_vector file_preds =
            context.parse_file(config.predicates_filename->c_str());

        for (expr e : file_preds) {
            predicates.push_back(re2aut.translate_predicate(e));
        }
    }
}

void ParikhProfile::find_new_predicates_from_automata_map(
    map<unsigned, shared_ptr<symbolic_automaton>> const &automata,
    set<unsigned> &pred_ids
) {
    for (auto item : automata) {
        auto aut = item.second;
        int pos = 0;
        for (unsigned s = 0; s <= aut->max_state(); s++) {
            if (aut->is_state(s)) {
                for (auto m : aut->get_moves_from(s)) {
                    if (m.t().has_value()) {
                        auto pred = m.t().value();
                        if (pred_ids.find(pred.id()) == pred_ids.end()) {
                            predicates.push_back(pred);
                            pred_ids.insert(pred.id());
                        }
                    }
                }
            }
        }
    }
}

void ParikhProfile::find_new_predicates_from_lits(std::set<unsigned> &pred_ids) {
    find_new_predicates_from_lits(in_re_checks, pred_ids);
    find_new_predicates_from_lits(not_in_re_checks, pred_ids);
    find_new_predicates_from_lits(seq_lens, pred_ids);
    find_new_predicates_from_lits(seq_equalities, pred_ids);
}

void ParikhProfile::find_new_predicates_from_lits(
    vector<expr> exprs,
    set<unsigned> &pred_ids
) {
    for (expr const &e : exprs)
        for (expr const &arg : e.args())
            find_new_predicates_from_lits(arg, pred_ids);
}

void ParikhProfile::find_new_predicates_from_lits(
    expr e,
    set<unsigned> &pred_ids
) {
    if (e.is_seq() && e.is_const() && e.is_string_value()) {
        for (unsigned char c : e.get_string()) {
            expr pred = re2aut.get_is_char(c);
            if (pred_ids.find(pred.id()) == pred_ids.end()) {
                predicates.push_back(pred);
                pred_ids.insert(pred.id());
            }
        }
    } else if (e.is_re()) {
        // ignore, these are already in the automata except for the arguments
        // to the in_range predicates, which we don't count
    } else {
        for (expr arg: e.args())
            find_new_predicates_from_lits(arg, pred_ids);
    }
}

/**
 * Helper function for exploring Z3 data structures
 */
static const char* decl_kind_to_string(const Z3_decl_kind& kind) {
    switch (kind) {
    case Z3_OP_TRUE: return "Z3_OP_TRUE";
    case Z3_OP_FALSE: return "Z3_OP_FALSE";
    case Z3_OP_EQ: return "Z3_OP_EQ";
    case Z3_OP_DISTINCT: return "Z3_OP_DISTINCT";
    case Z3_OP_ITE: return "Z3_OP_ITE";
    case Z3_OP_AND: return "Z3_OP_AND";
    case Z3_OP_OR: return "Z3_OP_OR";
    case Z3_OP_IFF: return "Z3_OP_IFF";
    case Z3_OP_XOR: return "Z3_OP_XOR";
    case Z3_OP_NOT: return "Z3_OP_NOT";
    case Z3_OP_IMPLIES: return "Z3_OP_IMPLIES";
    case Z3_OP_OEQ: return "Z3_OP_OEQ";
    case Z3_OP_ANUM: return "Z3_OP_ANUM";
    case Z3_OP_AGNUM: return "Z3_OP_AGNUM";
    case Z3_OP_LE: return "Z3_OP_LE";
    case Z3_OP_GE: return "Z3_OP_GE";
    case Z3_OP_LT: return "Z3_OP_LT";
    case Z3_OP_GT: return "Z3_OP_GT";
    case Z3_OP_ADD: return "Z3_OP_ADD";
    case Z3_OP_SUB: return "Z3_OP_SUB";
    case Z3_OP_UMINUS: return "Z3_OP_UMINUS";
    case Z3_OP_MUL: return "Z3_OP_MUL";
    case Z3_OP_DIV: return "Z3_OP_DIV";
    case Z3_OP_IDIV: return "Z3_OP_IDIV";
    case Z3_OP_REM: return "Z3_OP_REM";
    case Z3_OP_MOD: return "Z3_OP_MOD";
    case Z3_OP_TO_REAL: return "Z3_OP_TO_REAL";
    case Z3_OP_TO_INT: return "Z3_OP_TO_INT";
    case Z3_OP_IS_INT: return "Z3_OP_IS_INT";
    case Z3_OP_POWER: return "Z3_OP_POWER";
    case Z3_OP_STORE: return "Z3_OP_STORE";
    case Z3_OP_SELECT: return "Z3_OP_SELECT";
    case Z3_OP_CONST_ARRAY: return "Z3_OP_CONST_ARRAY";
    case Z3_OP_ARRAY_MAP: return "Z3_OP_ARRAY_MAP";
    case Z3_OP_ARRAY_DEFAULT: return "Z3_OP_ARRAY_DEFAULT";
    case Z3_OP_SET_UNION: return "Z3_OP_SET_UNION";
    case Z3_OP_SET_INTERSECT: return "Z3_OP_SET_INTERSECT";
    case Z3_OP_SET_DIFFERENCE: return "Z3_OP_SET_DIFFERENCE";
    case Z3_OP_SET_COMPLEMENT: return "Z3_OP_SET_COMPLEMENT";
    case Z3_OP_SET_SUBSET: return "Z3_OP_SET_SUBSET";
    case Z3_OP_AS_ARRAY: return "Z3_OP_AS_ARRAY";
    case Z3_OP_ARRAY_EXT: return "Z3_OP_ARRAY_EXT";
    case Z3_OP_SET_HAS_SIZE: return "Z3_OP_SET_HAS_SIZE";
    case Z3_OP_SET_CARD: return "Z3_OP_SET_CARD";
    case Z3_OP_BNUM: return "Z3_OP_BNUM";
    case Z3_OP_BIT1: return "Z3_OP_BIT1";
    case Z3_OP_BIT0: return "Z3_OP_BIT0";
    case Z3_OP_BNEG: return "Z3_OP_BNEG";
    case Z3_OP_BADD: return "Z3_OP_BADD";
    case Z3_OP_BSUB: return "Z3_OP_BSUB";
    case Z3_OP_BMUL: return "Z3_OP_BMUL";
    case Z3_OP_BSDIV: return "Z3_OP_BSDIV";
    case Z3_OP_BUDIV: return "Z3_OP_BUDIV";
    case Z3_OP_BSREM: return "Z3_OP_BSREM";
    case Z3_OP_BUREM: return "Z3_OP_BUREM";
    case Z3_OP_BSMOD: return "Z3_OP_BSMOD";
    case Z3_OP_BSDIV0: return "Z3_OP_BSDIV0";
    case Z3_OP_BUDIV0: return "Z3_OP_BUDIV0";
    case Z3_OP_BSREM0: return "Z3_OP_BSREM0";
    case Z3_OP_BUREM0: return "Z3_OP_BUREM0";
    case Z3_OP_BSMOD0: return "Z3_OP_BSMOD0";
    case Z3_OP_ULEQ: return "Z3_OP_ULEQ";
    case Z3_OP_SLEQ: return "Z3_OP_SLEQ";
    case Z3_OP_UGEQ: return "Z3_OP_UGEQ";
    case Z3_OP_SGEQ: return "Z3_OP_SGEQ";
    case Z3_OP_ULT: return "Z3_OP_ULT";
    case Z3_OP_SLT: return "Z3_OP_SLT";
    case Z3_OP_UGT: return "Z3_OP_UGT";
    case Z3_OP_SGT: return "Z3_OP_SGT";
    case Z3_OP_BAND: return "Z3_OP_BAND";
    case Z3_OP_BOR: return "Z3_OP_BOR";
    case Z3_OP_BNOT: return "Z3_OP_BNOT";
    case Z3_OP_BXOR: return "Z3_OP_BXOR";
    case Z3_OP_BNAND: return "Z3_OP_BNAND";
    case Z3_OP_BNOR: return "Z3_OP_BNOR";
    case Z3_OP_BXNOR: return "Z3_OP_BXNOR";
    case Z3_OP_CONCAT: return "Z3_OP_CONCAT";
    case Z3_OP_SIGN_EXT: return "Z3_OP_SIGN_EXT";
    case Z3_OP_ZERO_EXT: return "Z3_OP_ZERO_EXT";
    case Z3_OP_EXTRACT: return "Z3_OP_EXTRACT";
    case Z3_OP_REPEAT: return "Z3_OP_REPEAT";
    case Z3_OP_BREDOR: return "Z3_OP_BREDOR";
    case Z3_OP_BREDAND: return "Z3_OP_BREDAND";
    case Z3_OP_BCOMP: return "Z3_OP_BCOMP";
    case Z3_OP_BSHL: return "Z3_OP_BSHL";
    case Z3_OP_BLSHR: return "Z3_OP_BLSHR";
    case Z3_OP_BASHR: return "Z3_OP_BASHR";
    case Z3_OP_ROTATE_LEFT: return "Z3_OP_ROTATE_LEFT";
    case Z3_OP_ROTATE_RIGHT: return "Z3_OP_ROTATE_RIGHT";
    case Z3_OP_EXT_ROTATE_LEFT: return "Z3_OP_EXT_ROTATE_LEFT";
    case Z3_OP_EXT_ROTATE_RIGHT: return "Z3_OP_EXT_ROTATE_RIGHT";
    case Z3_OP_BIT2BOOL: return "Z3_OP_BIT2BOOL";
    case Z3_OP_INT2BV: return "Z3_OP_INT2BV";
    case Z3_OP_BV2INT: return "Z3_OP_BV2INT";
    case Z3_OP_CARRY: return "Z3_OP_CARRY";
    case Z3_OP_XOR3: return "Z3_OP_XOR3";
    case Z3_OP_BSMUL_NO_OVFL: return "Z3_OP_BSMUL_NO_OVFL";
    case Z3_OP_BUMUL_NO_OVFL: return "Z3_OP_BUMUL_NO_OVFL";
    case Z3_OP_BSMUL_NO_UDFL: return "Z3_OP_BSMUL_NO_UDFL";
    case Z3_OP_BSDIV_I: return "Z3_OP_BSDIV_I";
    case Z3_OP_BUDIV_I: return "Z3_OP_BUDIV_I";
    case Z3_OP_BSREM_I: return "Z3_OP_BSREM_I";
    case Z3_OP_BUREM_I: return "Z3_OP_BUREM_I";
    case Z3_OP_BSMOD_I: return "Z3_OP_BSMOD_I";
    case Z3_OP_PR_UNDEF: return "Z3_OP_PR_UNDEF";
    case Z3_OP_PR_TRUE: return "Z3_OP_PR_TRUE";
    case Z3_OP_PR_ASSERTED: return "Z3_OP_PR_ASSERTED";
    case Z3_OP_PR_GOAL: return "Z3_OP_PR_GOAL";
    case Z3_OP_PR_MODUS_PONENS: return "Z3_OP_PR_MODUS_PONENS";
    case Z3_OP_PR_REFLEXIVITY: return "Z3_OP_PR_REFLEXIVITY";
    case Z3_OP_PR_SYMMETRY: return "Z3_OP_PR_SYMMETRY";
    case Z3_OP_PR_TRANSITIVITY: return "Z3_OP_PR_TRANSITIVITY";
    case Z3_OP_PR_TRANSITIVITY_STAR: return "Z3_OP_PR_TRANSITIVITY_STAR";
    case Z3_OP_PR_MONOTONICITY: return "Z3_OP_PR_MONOTONICITY";
    case Z3_OP_PR_QUANT_INTRO: return "Z3_OP_PR_QUANT_INTRO";
    case Z3_OP_PR_BIND: return "Z3_OP_PR_BIND";
    case Z3_OP_PR_DISTRIBUTIVITY: return "Z3_OP_PR_DISTRIBUTIVITY";
    case Z3_OP_PR_AND_ELIM: return "Z3_OP_PR_AND_ELIM";
    case Z3_OP_PR_NOT_OR_ELIM: return "Z3_OP_PR_NOT_OR_ELIM";
    case Z3_OP_PR_REWRITE: return "Z3_OP_PR_REWRITE";
    case Z3_OP_PR_REWRITE_STAR: return "Z3_OP_PR_REWRITE_STAR";
    case Z3_OP_PR_PULL_QUANT: return "Z3_OP_PR_PULL_QUANT";
    case Z3_OP_PR_PUSH_QUANT: return "Z3_OP_PR_PUSH_QUANT";
    case Z3_OP_PR_ELIM_UNUSED_VARS: return "Z3_OP_PR_ELIM_UNUSED_VARS";
    case Z3_OP_PR_DER: return "Z3_OP_PR_DER";
    case Z3_OP_PR_QUANT_INST: return "Z3_OP_PR_QUANT_INST";
    case Z3_OP_PR_HYPOTHESIS: return "Z3_OP_PR_HYPOTHESIS";
    case Z3_OP_PR_LEMMA: return "Z3_OP_PR_LEMMA";
    case Z3_OP_PR_UNIT_RESOLUTION: return "Z3_OP_PR_UNIT_RESOLUTION";
    case Z3_OP_PR_IFF_TRUE: return "Z3_OP_PR_IFF_TRUE";
    case Z3_OP_PR_IFF_FALSE: return "Z3_OP_PR_IFF_FALSE";
    case Z3_OP_PR_COMMUTATIVITY: return "Z3_OP_PR_COMMUTATIVITY";
    case Z3_OP_PR_DEF_AXIOM: return "Z3_OP_PR_DEF_AXIOM";
    case Z3_OP_PR_ASSUMPTION_ADD: return "Z3_OP_PR_ASSUMPTION_ADD";
    case Z3_OP_PR_LEMMA_ADD: return "Z3_OP_PR_LEMMA_ADD";
    case Z3_OP_PR_REDUNDANT_DEL: return "Z3_OP_PR_REDUNDANT_DEL";
    case Z3_OP_PR_CLAUSE_TRAIL: return "Z3_OP_PR_CLAUSE_TRAIL";
    case Z3_OP_PR_DEF_INTRO: return "Z3_OP_PR_DEF_INTRO";
    case Z3_OP_PR_APPLY_DEF: return "Z3_OP_PR_APPLY_DEF";
    case Z3_OP_PR_IFF_OEQ: return "Z3_OP_PR_IFF_OEQ";
    case Z3_OP_PR_NNF_POS: return "Z3_OP_PR_NNF_POS";
    case Z3_OP_PR_NNF_NEG: return "Z3_OP_PR_NNF_NEG";
    case Z3_OP_PR_SKOLEMIZE: return "Z3_OP_PR_SKOLEMIZE";
    case Z3_OP_PR_MODUS_PONENS_OEQ: return "Z3_OP_PR_MODUS_PONENS_OEQ";
    case Z3_OP_PR_TH_LEMMA: return "Z3_OP_PR_TH_LEMMA";
    case Z3_OP_PR_HYPER_RESOLVE: return "Z3_OP_PR_HYPER_RESOLVE";
    case Z3_OP_RA_STORE: return "Z3_OP_RA_STORE";
    case Z3_OP_RA_EMPTY: return "Z3_OP_RA_EMPTY";
    case Z3_OP_RA_IS_EMPTY: return "Z3_OP_RA_IS_EMPTY";
    case Z3_OP_RA_JOIN: return "Z3_OP_RA_JOIN";
    case Z3_OP_RA_UNION: return "Z3_OP_RA_UNION";
    case Z3_OP_RA_WIDEN: return "Z3_OP_RA_WIDEN";
    case Z3_OP_RA_PROJECT: return "Z3_OP_RA_PROJECT";
    case Z3_OP_RA_FILTER: return "Z3_OP_RA_FILTER";
    case Z3_OP_RA_NEGATION_FILTER: return "Z3_OP_RA_NEGATION_FILTER";
    case Z3_OP_RA_RENAME: return "Z3_OP_RA_RENAME";
    case Z3_OP_RA_COMPLEMENT: return "Z3_OP_RA_COMPLEMENT";
    case Z3_OP_RA_SELECT: return "Z3_OP_RA_SELECT";
    case Z3_OP_RA_CLONE: return "Z3_OP_RA_CLONE";
    case Z3_OP_FD_CONSTANT: return "Z3_OP_FD_CONSTANT";
    case Z3_OP_FD_LT: return "Z3_OP_FD_LT";
    case Z3_OP_SEQ_UNIT: return "Z3_OP_SEQ_UNIT";
    case Z3_OP_SEQ_EMPTY: return "Z3_OP_SEQ_EMPTY";
    case Z3_OP_SEQ_CONCAT: return "Z3_OP_SEQ_CONCAT";
    case Z3_OP_SEQ_PREFIX: return "Z3_OP_SEQ_PREFIX";
    case Z3_OP_SEQ_SUFFIX: return "Z3_OP_SEQ_SUFFIX";
    case Z3_OP_SEQ_CONTAINS: return "Z3_OP_SEQ_CONTAINS";
    case Z3_OP_SEQ_EXTRACT: return "Z3_OP_SEQ_EXTRACT";
    case Z3_OP_SEQ_REPLACE: return "Z3_OP_SEQ_REPLACE";
    case Z3_OP_SEQ_REPLACE_RE: return "Z3_OP_SEQ_REPLACE_RE";
    case Z3_OP_SEQ_REPLACE_RE_ALL: return "Z3_OP_SEQ_REPLACE_RE_ALL";
    case Z3_OP_SEQ_REPLACE_ALL: return "Z3_OP_SEQ_REPLACE_ALL";
    case Z3_OP_SEQ_AT: return "Z3_OP_SEQ_AT";
    case Z3_OP_SEQ_NTH: return "Z3_OP_SEQ_NTH";
    case Z3_OP_SEQ_LENGTH: return "Z3_OP_SEQ_LENGTH";
    case Z3_OP_SEQ_INDEX: return "Z3_OP_SEQ_INDEX";
    case Z3_OP_SEQ_LAST_INDEX: return "Z3_OP_SEQ_LAST_INDEX";
    case Z3_OP_SEQ_TO_RE: return "Z3_OP_SEQ_TO_RE";
    case Z3_OP_SEQ_IN_RE: return "Z3_OP_SEQ_IN_RE";
    case Z3_OP_STR_TO_INT: return "Z3_OP_STR_TO_INT";
    case Z3_OP_INT_TO_STR: return "Z3_OP_INT_TO_STR";
    case Z3_OP_UBV_TO_STR: return "Z3_OP_UBV_TO_STR";
    case Z3_OP_SBV_TO_STR: return "Z3_OP_SBV_TO_STR";
    case Z3_OP_STR_TO_CODE: return "Z3_OP_STR_TO_CODE";
    case Z3_OP_STR_FROM_CODE: return "Z3_OP_STR_FROM_CODE";
    case Z3_OP_STRING_LT: return "Z3_OP_STRING_LT";
    case Z3_OP_STRING_LE: return "Z3_OP_STRING_LE";
    case Z3_OP_RE_PLUS: return "Z3_OP_RE_PLUS";
    case Z3_OP_RE_STAR: return "Z3_OP_RE_STAR";
    case Z3_OP_RE_OPTION: return "Z3_OP_RE_OPTION";
    case Z3_OP_RE_CONCAT: return "Z3_OP_RE_CONCAT";
    case Z3_OP_RE_UNION: return "Z3_OP_RE_UNION";
    case Z3_OP_RE_RANGE: return "Z3_OP_RE_RANGE";
    case Z3_OP_RE_DIFF: return "Z3_OP_RE_DIFF";
    case Z3_OP_RE_INTERSECT: return "Z3_OP_RE_INTERSECT";
    case Z3_OP_RE_LOOP: return "Z3_OP_RE_LOOP";
    case Z3_OP_RE_POWER: return "Z3_OP_RE_POWER";
    case Z3_OP_RE_COMPLEMENT: return "Z3_OP_RE_COMPLEMENT";
    case Z3_OP_RE_EMPTY_SET: return "Z3_OP_RE_EMPTY_SET";
    case Z3_OP_RE_FULL_SET: return "Z3_OP_RE_FULL_SET";
    case Z3_OP_RE_FULL_CHAR_SET: return "Z3_OP_RE_FULL_CHAR_SET";
    case Z3_OP_RE_OF_PRED: return "Z3_OP_RE_OF_PRED";
    case Z3_OP_RE_REVERSE: return "Z3_OP_RE_REVERSE";
    case Z3_OP_RE_DERIVATIVE: return "Z3_OP_RE_DERIVATIVE";
    case Z3_OP_CHAR_CONST: return "Z3_OP_CHAR_CONST";
    case Z3_OP_CHAR_LE: return "Z3_OP_CHAR_LE";
    case Z3_OP_CHAR_TO_INT: return "Z3_OP_CHAR_TO_INT";
    case Z3_OP_CHAR_TO_BV: return "Z3_OP_CHAR_TO_BV";
    case Z3_OP_CHAR_FROM_BV: return "Z3_OP_CHAR_FROM_BV";
    case Z3_OP_CHAR_IS_DIGIT: return "Z3_OP_CHAR_IS_DIGIT";
    case Z3_OP_LABEL: return "Z3_OP_LABEL";
    case Z3_OP_LABEL_LIT: return "Z3_OP_LABEL_LIT";
    case Z3_OP_DT_CONSTRUCTOR: return "Z3_OP_DT_CONSTRUCTOR";
    case Z3_OP_DT_RECOGNISER: return "Z3_OP_DT_RECOGNISER";
    case Z3_OP_DT_IS: return "Z3_OP_DT_IS";
    case Z3_OP_DT_ACCESSOR: return "Z3_OP_DT_ACCESSOR";
    case Z3_OP_DT_UPDATE_FIELD: return "Z3_OP_DT_UPDATE_FIELD";
    case Z3_OP_PB_AT_MOST: return "Z3_OP_PB_AT_MOST";
    case Z3_OP_PB_AT_LEAST: return "Z3_OP_PB_AT_LEAST";
    case Z3_OP_PB_LE: return "Z3_OP_PB_LE";
    case Z3_OP_PB_GE: return "Z3_OP_PB_GE";
    case Z3_OP_PB_EQ: return "Z3_OP_PB_EQ";
    case Z3_OP_SPECIAL_RELATION_LO: return "Z3_OP_SPECIAL_RELATION_LO";
    case Z3_OP_SPECIAL_RELATION_PO: return "Z3_OP_SPECIAL_RELATION_PO";
    case Z3_OP_SPECIAL_RELATION_PLO: return "Z3_OP_SPECIAL_RELATION_PLO";
    case Z3_OP_SPECIAL_RELATION_TO: return "Z3_OP_SPECIAL_RELATION_TO";
    case Z3_OP_SPECIAL_RELATION_TC: return "Z3_OP_SPECIAL_RELATION_TC";
    case Z3_OP_SPECIAL_RELATION_TRC: return "Z3_OP_SPECIAL_RELATION_TRC";
    case Z3_OP_FPA_RM_NEAREST_TIES_TO_EVEN: return "Z3_OP_FPA_RM_NEAREST_TIES_TO_EVEN";
    case Z3_OP_FPA_RM_NEAREST_TIES_TO_AWAY: return "Z3_OP_FPA_RM_NEAREST_TIES_TO_AWAY";
    case Z3_OP_FPA_RM_TOWARD_POSITIVE: return "Z3_OP_FPA_RM_TOWARD_POSITIVE";
    case Z3_OP_FPA_RM_TOWARD_NEGATIVE: return "Z3_OP_FPA_RM_TOWARD_NEGATIVE";
    case Z3_OP_FPA_RM_TOWARD_ZERO: return "Z3_OP_FPA_RM_TOWARD_ZERO";
    case Z3_OP_FPA_NUM: return "Z3_OP_FPA_NUM";
    case Z3_OP_FPA_PLUS_INF: return "Z3_OP_FPA_PLUS_INF";
    case Z3_OP_FPA_MINUS_INF: return "Z3_OP_FPA_MINUS_INF";
    case Z3_OP_FPA_NAN: return "Z3_OP_FPA_NAN";
    case Z3_OP_FPA_PLUS_ZERO: return "Z3_OP_FPA_PLUS_ZERO";
    case Z3_OP_FPA_MINUS_ZERO: return "Z3_OP_FPA_MINUS_ZERO";
    case Z3_OP_FPA_ADD: return "Z3_OP_FPA_ADD";
    case Z3_OP_FPA_SUB: return "Z3_OP_FPA_SUB";
    case Z3_OP_FPA_NEG: return "Z3_OP_FPA_NEG";
    case Z3_OP_FPA_MUL: return "Z3_OP_FPA_MUL";
    case Z3_OP_FPA_DIV: return "Z3_OP_FPA_DIV";
    case Z3_OP_FPA_REM: return "Z3_OP_FPA_REM";
    case Z3_OP_FPA_ABS: return "Z3_OP_FPA_ABS";
    case Z3_OP_FPA_MIN: return "Z3_OP_FPA_MIN";
    case Z3_OP_FPA_MAX: return "Z3_OP_FPA_MAX";
    case Z3_OP_FPA_FMA: return "Z3_OP_FPA_FMA";
    case Z3_OP_FPA_SQRT: return "Z3_OP_FPA_SQRT";
    case Z3_OP_FPA_ROUND_TO_INTEGRAL: return "Z3_OP_FPA_ROUND_TO_INTEGRAL";
    case Z3_OP_FPA_EQ: return "Z3_OP_FPA_EQ";
    case Z3_OP_FPA_LT: return "Z3_OP_FPA_LT";
    case Z3_OP_FPA_GT: return "Z3_OP_FPA_GT";
    case Z3_OP_FPA_LE: return "Z3_OP_FPA_LE";
    case Z3_OP_FPA_GE: return "Z3_OP_FPA_GE";
    case Z3_OP_FPA_IS_NAN: return "Z3_OP_FPA_IS_NAN";
    case Z3_OP_FPA_IS_INF: return "Z3_OP_FPA_IS_INF";
    case Z3_OP_FPA_IS_ZERO: return "Z3_OP_FPA_IS_ZERO";
    case Z3_OP_FPA_IS_NORMAL: return "Z3_OP_FPA_IS_NORMAL";
    case Z3_OP_FPA_IS_SUBNORMAL: return "Z3_OP_FPA_IS_SUBNORMAL";
    case Z3_OP_FPA_IS_NEGATIVE: return "Z3_OP_FPA_IS_NEGATIVE";
    case Z3_OP_FPA_IS_POSITIVE: return "Z3_OP_FPA_IS_POSITIVE";
    case Z3_OP_FPA_FP: return "Z3_OP_FPA_FP";
    case Z3_OP_FPA_TO_FP: return "Z3_OP_FPA_TO_FP";
    case Z3_OP_FPA_TO_FP_UNSIGNED: return "Z3_OP_FPA_TO_FP_UNSIGNED";
    case Z3_OP_FPA_TO_UBV: return "Z3_OP_FPA_TO_UBV";
    case Z3_OP_FPA_TO_SBV: return "Z3_OP_FPA_TO_SBV";
    case Z3_OP_FPA_TO_REAL: return "Z3_OP_FPA_TO_REAL";
    case Z3_OP_FPA_TO_IEEE_BV: return "Z3_OP_FPA_TO_IEEE_BV";
    case Z3_OP_FPA_BVWRAP: return "Z3_OP_FPA_BVWRAP";
    case Z3_OP_FPA_BV2RM: return "Z3_OP_FPA_BV2RM";
    case Z3_OP_INTERNAL: return "Z3_OP_INTERNAL";
    case Z3_OP_RECURSIVE: return "Z3_OP_RECURSIVE";
    case Z3_OP_UNINTERPRETED: return "Z3_OP_UNINTERPRETED";
    }
    return "Unknown";
}
