
#ifndef __PARIKH_PROFILE_H__
#define __PARIKH_PROFILE_H__

#include <exception>
#include <map>
#include <set>
#include <vector>

#include "z3++.h"

#include "re2automaton.h"
#include "symbolic_automaton.h"

namespace sym_parikh{
    class ParikhProfileException : public std::exception {
        public:
            ParikhProfileException(std::string const &reason)
                    : reason(reason) { }

            virtual const char* what() const throw() {
                return reason.c_str();
            }

        private:
            std::string const reason;
    };

    struct ParikhProfileConfig {
        std::shared_ptr<std::string> predicates_filename;
        bool preds_from_lits;

        ParikhProfileConfig() : ParikhProfileConfig(nullptr, false) { }

        /**
         * Build a config
         *
         * @param predicates_filename a file to read preds from or null if to
         * be inferred
         * @param preds_from_lits whether to infer predicates from chars in
         * string literals
         */
        ParikhProfileConfig(
            std::shared_ptr<std::string> predicates_filename,
            bool preds_from_lits
        ) : predicates_filename(predicates_filename),
            preds_from_lits(preds_from_lits) { }
    };


    /**
     * Extracts data from an expression to aid Parikh pre-processing
     *
     * E.g. stuff like the list of string variables and the expressions
     * over strings.
     *
     * Expressions must be in negation normal form, else results will be
     * unreliable
     */
    class ParikhProfile {
        public:
            ParikhProfile(
                z3::context &context,
                z3::expr_vector const &formulas
            ) : ParikhProfile(context, formulas, ParikhProfileConfig()) { }


            ParikhProfile(
                z3::context &context,
                z3::expr_vector const &formulas,
                ParikhProfileConfig config
            ) : config(config),
                    context(context),
                    re2aut(context),
                    formulas(formulas),
                    seq_vars(),
                    in_re_checks(),
                    not_in_re_checks(),
                    positive_automata(),
                    negative_automata(),
                    predicates(),
                    has_length(false),
                    seen_expr_ids() {
                extract_data(formulas);
                make_automata();
                find_predicates();
            }

            /**
             * Get all seq vars in formula
             */
            std::vector<z3::expr> const get_seq_vars() const {
                return seq_vars;
            }

            /**
             * Get all positive (str.in_re x re) expressions in formula
             *
             * Note, x may be something like (str.++ x y "a" z)
             */
            std::vector<z3::expr> const get_in_re_checks() const {
                return in_re_checks;
            }

            /**
             * Get all (not (str.in_re x re)) expressions in formula
             *
             * Note, x may be something like (str.++ x y "a" z)
             */
            std::vector<z3::expr> const get_not_in_re_checks() const {
                return not_in_re_checks;
            }

            /**
             * Get automaton for an expression in an re check or re equality
             */
            std::shared_ptr<symbolic_automaton> get_positive_automaton(
                z3::expr const &re
            ) {
                return positive_automata.find(re.id())->second;
            }

            /**
             * Get automaton for expression in a neg re check or neg re equality
             */
            std::shared_ptr<symbolic_automaton> get_negative_automaton(
                z3::expr const &re
            ) {
                return negative_automata.find(re.id())->second;
            }

            /**
             * Get all (str.len x) calls
             *
             * Note, x may be something like (str.++ x y "a" z)
             */
            std::vector<z3::expr> const get_seq_lens() const {
                return seq_lens;
            }

            /**
             * Get all (= string1 string2), (str.prefix string1 string2) etc
             *
             * string1 and string2 may be e.g. (str.++ x y "z")
             *
             * Supports =, str.prefixof, str.suffixof, str.contains
             */
            std::vector<z3::expr> const get_seq_equalities() const {
                return seq_equalities;
            }

            /**
             * Get all (= re1 re2) for regex equality
             */
            std::vector<z3::expr> const get_re_equalities() const {
                return re_equalities;
            }

            /**
             * Get all (not (= re1 re2)) for regex not inequality
             */
            std::vector<z3::expr> const get_re_notequalities() const {
                return re_notequalities;
            }

            /**
             * Predicates appearing in the re automata
             */
            std::vector<z3::expr> const get_predicates() const {
                return predicates;
            }

            std::shared_ptr<symbolic_automaton> get_positive_automaton(
                z3::expr const &re
            ) const {
                auto v = positive_automata.find(re.id());
                if (v == positive_automata.end()) return NULL;
                else return v->second;
            }

            std::shared_ptr<symbolic_automaton> get_negative_automaton(
                z3::expr const &re
            ) const {
                auto v = negative_automata.find(re.id());
                if (v == negative_automata.end()) return NULL;
                else return v->second;
            }

        private:
            ParikhProfileConfig config;
            z3::context &context;
            re2automaton re2aut;

            z3::expr_vector const formulas;
            std::vector<z3::expr> seq_vars;
            std::vector<z3::expr> in_re_checks;
            std::vector<z3::expr> not_in_re_checks;
            std::vector<z3::expr> seq_lens;
            std::vector<z3::expr> seq_equalities;
            std::vector<z3::expr> re_equalities;
            std::vector<z3::expr> re_notequalities;
            std::map<unsigned, std::shared_ptr<symbolic_automaton>>
                positive_automata;
            std::map<unsigned, std::shared_ptr<symbolic_automaton>>
                negative_automata;
            std::vector<z3::expr> predicates;
            bool has_length;

            // the expression ids already in one of the data sets above
            std::set<unsigned> seen_expr_ids;

            /**
             * Extract needed data from formulas
             */
            void extract_data(z3::expr_vector const &formula) {
                for (z3::expr f : formulas)
                    extract_data(f);
            }

            /**
             * Extract needed data from formulas
             */
            void extract_data(z3::expr const &formula);

            /**
             * Make automata out of re checks in re_checks
             */
            void make_automata();

            /**
             * Fill the predicates list with predicates from automata
             */
            void find_predicates();

            /**
             * Fill the predicates list with predicates from automata
             *
             * Ignores predicates already in pred_ids, adds new ones to
             * pred_ids too
             */
            void find_new_predicates_from_automata_map(
                std::map<
                    unsigned,
                    std::shared_ptr<symbolic_automaton>
                > const &automata,
                std::set<unsigned> &pred_ids
            );

            /**
             * Fill the predicates list with predicates from string literals
             *
             * Looks through string expressions for string literatals. Adds
             * predicates for each char found in a literal.
             *
             * E.g. (x + "ab" + y) in re would add =a and =b.
             *
             * Ignores predicates already in pred_ids.
             */
            void find_new_predicates_from_lits(std::set<unsigned> &pred_ids);

            /**
             * Finds new preds from string expressions in exprs in list
             */
            void find_new_predicates_from_lits(
                std::vector<z3::expr> exprs,
                std::set<unsigned> &pred_ids
            );

            /**
             * Finds new preds from expression
             */
            void find_new_predicates_from_lits(
                z3::expr const e,
                std::set<unsigned> &pred_ids
            );

    };
}

#endif
