
// for easy debugging
#include <iostream>
#include <sstream>

#include "z3++.h"

#include "symbolic_automaton.h"
#include "re2automaton.h"
#include "z3types/regex.h"
#include "z3types/zstring.h"

using namespace std;
using namespace sym_parikh;
using namespace z3;

string const re2automaton::free_variable = "_x_";

shared_ptr<symbolic_automaton> re2automaton::translate(expr const &re) {
    shared_ptr<symbolic_automaton> aut(re2aut(re));
    aut->compress();
    return aut;
}

shared_ptr<symbolic_automaton> re2automaton::translate_negative(
    expr const &re
) {
    shared_ptr<symbolic_automaton> aut(re2aut(re));
    aut->compress();
    shared_ptr<symbolic_automaton> neg(
        symbolic_automaton::mk_complement(context, *aut)
    );
    return neg;
}

static unique_ptr<symbolic_automaton> unique(symbolic_automaton *a) {
    return unique_ptr<symbolic_automaton>(a);
}

/**
 * Convert re to automaton without compression
 *
 * Borrowed as much as possible from z3 code
 */
symbolic_automaton* re2automaton::re2aut(expr const &re) {
    std::vector<expr> args;
    unique_ptr<symbolic_automaton> a, b;
    shared_ptr<symbolic_automaton> result;
    unsigned lo, hi;
    if (regex::is_to_re(re, args)) {
        return seq2aut(args[0]);
    }
    else if (regex::is_concat(re, args)) {
        if (args.size() == 0)
            return symbolic_automaton::mk_none();
        a = unique(re2aut(args[0]));
        for (int i = 1; i < args.size(); i++) {
            b = unique(re2aut(args[i]));
            a = unique(symbolic_automaton::mk_concat(*a, *b));
        }
        return a.release();
    }
    else if (regex::is_union(re, args)) {
        if (args.size() == 0)
            return symbolic_automaton::mk_none();
        a = unique(re2aut(args[0]));
        for (int i = 1; i < args.size(); i++) {
            b = unique(re2aut(args[i]));
            a = unique(symbolic_automaton::mk_union(*a, *b));
        }
        return a.release();
    }
    else if (
        regex::is_star(re, args) &&
        (a = unique(re2aut(args[0])))
    ) {
        a->add_final_to_init_moves();
        a->add_init_to_final_states();
        return a.release();
    }
    else if (
        regex::is_plus(re, args) &&
        (a = unique(re2aut(args[0])))
    ) {
        a->add_final_to_init_moves();
        return a.release();
    }
    else if (
        regex::is_opt(re, args) &&
        (a = unique(re2aut(args[0])))
    ) {
        a = unique(symbolic_automaton::mk_opt(*a));
        return a.release();
    }
    else if (regex::is_range(re, args)) {
        int lo = get_char(args[0]);
        int hi = get_char(args[1]);
        return new symbolic_automaton(get_in_range(lo, hi));
    }
    else if (
        regex::is_complement(re, args) &&
        (a = unique(re2aut(args[0])))
    ) {
        return symbolic_automaton::mk_complement(context, *a);
    }
    else if (
        regex::is_loop(re, args, lo, hi) &&
        (a = unique(re2aut(args[0])))
    ) {
        unique_ptr<symbolic_automaton> eps(symbolic_automaton::mk_epsilon());
        b = unique(symbolic_automaton::mk_epsilon());
        while (hi > lo) {
            unique_ptr<symbolic_automaton> c = unique(
                symbolic_automaton::mk_concat(*a, *b)
            );
            b = unique(symbolic_automaton::mk_union(*eps, *c));
            --hi;
        }
        while (lo > 0) {
            b = unique(symbolic_automaton::mk_concat(*a, *b));
            --lo;
        }
        return b.release();
    }
    else if (
        regex::is_loop(re, args, lo) &&
        (a = unique(re2aut(args[0])))
    ) {
        b = unique(symbolic_automaton::clone(*a));
        b->add_final_to_init_moves();
        b->add_init_to_final_states();
        while (lo > 0) {
            b = unique(symbolic_automaton::mk_concat(*a, *b));
            --lo;
        }
        return b.release();
    }
    else if (regex::is_empty(re)) {
        return new symbolic_automaton();
    }
    else if (regex::is_full_seq(re)) {
        expr tt = context.bool_val(true);
        return symbolic_automaton::mk_loop(tt);
    }
    else if (regex::is_full_char(re)) {
        expr tt = context.bool_val(true);
        return new symbolic_automaton(tt);
    }
    else if (
        regex::is_intersection(re, args) &&
        (a = unique(re2aut(args[0]))) &&
        (b = unique(re2aut(args[1])))
    ) {
        symbolic_automaton* r =
            symbolic_automaton::mk_product(context, *a, *b);
        return r;
    }
    else if (regex::is_none(re)) {
        symbolic_automaton* r = symbolic_automaton::mk_none();
        return r;
    }
    else {
        assert(false); // fail fast?
    }

    return nullptr;
}

/**
 * Convert seq to automaton without compression
 *
 * Borrowed as much as possible from z3 code
 */
symbolic_automaton* re2automaton::seq2aut(expr const &e) {
    std::vector<expr> args;
    std::unique_ptr<symbolic_automaton> a, b;
    std::string s;

    if (
        zstring::is_concat(e, args) &&
        (a = unique(seq2aut(args[0]))) &&
        (b = unique(seq2aut(args[1])))
    ) {
        return symbolic_automaton::mk_concat(*a, *b);
    }
    else if (zstring::is_unit(e, args)) {
        return new symbolic_automaton(get_is_char(get_char(args[0])));
    }
    else if (zstring::is_empty(e)) {
        return symbolic_automaton::mk_epsilon();
    }
    else if (zstring::is_string(e, s)) {
        unsigned init = 0;
        symbolic_automaton::moves mvs;
        unsigned k = 0;
        unsigned state = 0;
        while (k < s.length()) {
            int c;
            k = chomp_char(s, k, c);
            mvs.insert(
                symbolic_automaton::move(state, state+1, get_is_char(c))
            );
            state += 1;
        }
        std::vector<unsigned> final;
        final.push_back(state);
        return new symbolic_automaton(init, final, mvs);
    }
    else if (zstring::is_param(e)) {
        Z3_ast c = Z3_mk_string_to_code(context, e);
        Z3_ast p = Z3_mk_string_from_code(context, c);
        expr te(context, c);
        symbolic_automaton chraut(get_is_expr(context, te));
        return new symbolic_automaton(get_is_expr(context, te));
    }
    return nullptr;
}

unsigned re2automaton::chomp_char(
    string const &s,
    unsigned pos,
    int &out
) {
    // need at least 5 chars for \u{.}
    if (pos + 5 > s.length()) {
        out = (int) s[pos];
        return pos + 1;
    }

    if (
        s[pos] != '\\' ||
        s[pos + 1] != 'u' ||
        s[pos + 2] != '{'
    ) {
        out = (int) s[pos];
        return pos + 1;
    }

    string char_code = "";
    unsigned char_pos = pos + 3;
    while (char_pos < s.length() && s[char_pos] != '}') {
        if (
            ('0' <= s[char_pos] && s[char_pos] <= '9') ||
            ('a' <= s[char_pos] && s[char_pos] <= 'f') ||
            ('A' <= s[char_pos] && s[char_pos] <= 'F')
        ) {
            char_code = char_code + s[char_pos];
            char_pos += 1;
        } else {
            out = (int) s[pos];
            return pos + 1;
        }
    }

    if (s[char_pos] == '}') {
        stringstream ss;
        out = stoi(char_code, nullptr, 16);
        return char_pos + 1;
    } else {
        out = (int) s[pos];
        return pos + 1;
    }
}

int re2automaton::get_char(expr const &c) {
    std::vector<expr> args;
    std::string s;

    if (zstring::is_unit(c, args)) {
        return get_char(args[0]);
    } else if (zstring::is_string(c, s)) {
        int chr;
        unsigned end = chomp_char(s, 0, chr);
        if (end >= s.length())
            return chr;
    }

    ostringstream ss;
    ss << "Cannot convert expression to char " << c;
    throw new RE2AutomatonException(ss.str());
}

expr re2automaton::translate_predicate(expr const& pred) {
    expr_vector old_exprs(context);
    expr_vector new_exprs(context);
    set<unsigned> seen_ids;

    predicate_extract_replaceable_exprs(
        pred,
        old_exprs,
        new_exprs,
        seen_ids
    );

    expr pred_copy = pred;
    return pred_copy.substitute(old_exprs, new_exprs);
}

void re2automaton::predicate_extract_replaceable_exprs(
    expr const &pred,
    expr_vector &old_exprs,
    expr_vector &new_exprs,
    set<unsigned> &seen_ids
) {
    if (seen_ids.find(pred.id()) != seen_ids.end())
        return;

    seen_ids.insert(pred.id());

    switch (pred.decl().decl_kind()) {
    case Z3_OP_SEQ_IN_RE: {
        if (pred.args().size() != 2) {
            ostringstream ss;
            ss << "Unexpected number of args to " << pred;
            throw new RE2AutomatonException(ss.str());
        }
        expr lhs = pred.args()[0];
        expr rhs = pred.args()[1];

        if (!is_string_var(lhs)) {
            ostringstream ss;
            ss << "Unexpected LHS of " << pred;
            throw new RE2AutomatonException(ss.str());
        }

        vector<expr> re_args;
        if (regex::is_range(rhs, re_args)) {
            int lo = get_char(re_args[0]);
            int hi = get_char(re_args[1]);
            old_exprs.push_back(pred);
            new_exprs.push_back(get_in_range(lo, hi));
        } else if (regex::is_to_re(rhs, re_args)) {
            expr chr = re_args[0];
            vector<expr> chr_args;
            string s;
            if (zstring::is_unit(chr, chr_args)) {
                old_exprs.push_back(pred);
                new_exprs.push_back(get_is_char(get_char(chr_args[0])));
            } else if (zstring::is_string(chr, s)) {
                if (s.length() != 1) {
                    ostringstream ss;
                    ss << "Only 1-char strings allowed " << pred;
                    throw new RE2AutomatonException(ss.str());
                }
                old_exprs.push_back(pred);
                new_exprs.push_back(get_is_char(s[0]));
            } else {
                ostringstream ss;
                ss << "Unrecognised in_re predicate " << pred;
                throw new RE2AutomatonException(ss.str());
            }
        }
        break;
    }
    case Z3_OP_EQ: {
        if (pred.args().size() != 2) {
            ostringstream ss;
            ss << "Unexpected number of predicate arguments " << pred;
            throw new RE2AutomatonException(ss.str());
        }

        expr lhs = pred.args()[0];
        expr rhs = pred.args()[1];

        if (is_string_var(lhs)) {
            old_exprs.push_back(pred);
            new_exprs.push_back(get_is_char(get_char(rhs)));
        } else if (is_string_var(rhs)) {
            old_exprs.push_back(pred);
            new_exprs.push_back(get_is_char(get_char(lhs)));
        } else {
            ostringstream ss;
            ss << "Equalities must have a string var: " << pred;
            throw new RE2AutomatonException(ss.str());
        }

        break;
    }
    // allowed list of other operators
    case Z3_OP_DISTINCT:
    case Z3_OP_ITE:
    case Z3_OP_AND:
    case Z3_OP_OR:
    case Z3_OP_IFF:
    case Z3_OP_NOT:
    case Z3_OP_IMPLIES:
        for (expr const &arg : pred.args()) {
            predicate_extract_replaceable_exprs(
                arg,
                old_exprs,
                new_exprs,
                seen_ids
            );
        }
        break;
    default:
        ostringstream ss;
        ss << "Unhandled input predicate " << pred;
        throw new RE2AutomatonException(ss.str());
    }
}

