
#ifndef __RE_2_AUTOMATON_H__
#define __RE_2_AUTOMATON_H__

#include <set>

#include "z3++.h"

#include "symbolic_automaton.h"

namespace sym_parikh {
    class RE2AutomatonException : public std::exception {
        public:
            RE2AutomatonException(std::string const &reason)
                    : reason(reason) { }

            virtual const char* what() const throw() {
                return reason.c_str();
            }

        private:
            std::string const reason;
    };

    class re2automaton {
        private:
            z3::context &context;

        public:
            /**
             * Build translator using context.
             *
             * Keeps reference to context, so don't use after context
             * out of scope.
             */
            re2automaton(z3::context &context)
                : context(context) { }

            /**
             * The free variable used for the transition predicates
             *
             * E.g. (str.in_re x (re.range "a" "b")). This is a bit of a
             * temporary solution pending further discussions.
             */
            static std::string const free_variable;

            /**
             * Make the free variable for the given context
             */
            static z3::expr get_free_variable(z3::context &context) {
                return context.int_const(free_variable.c_str());
            }

            /**
             * Get predicate asserting free var is a char for context
             */
            static z3::expr get_is_char(z3::context &context, char c) {
                return get_free_variable(context) == context.int_val((int) c);
            }

            /**
             * Get predicate asserting free var is a char for context
             */
            static z3::expr get_is_char(z3::context &context, int c) {
                return get_free_variable(context) == context.int_val(c);
            }

            /**
             * Get predicate asserting free var is a char for context
             */
            static z3::expr get_is_expr(z3::context &context, z3::expr const &e) {
                return get_free_variable(context) == e;
            }

            /**
             * Make free variable for object context
             */
            z3::expr get_free_variable() {
                return get_free_variable(context);
            }

            /**
             * Convert re into a symbolic automaton
             *
             * Needs the context as it does SAT checks when complementing
             * symbolic automata, and constructs some new formulas for the
             * transition predicates.
             *
             * @param context the context of z3 for re
             * @param re regular expression expected to be a z3 regex
             */
            std::shared_ptr<symbolic_automaton> translate(z3::expr const &re);

            /**
             * Convert re into a negative symbolic automaton
             *
             * @param context the context of z3 for re
             * @param re regular expression expected to be a z3 regex
             */
            std::shared_ptr<symbolic_automaton> translate_negative(
                z3::expr const &re
            );

            /**
             * Get predicate asserting free var is a char
             */
            z3::expr get_is_char(unsigned char c) {
                return get_is_char(context, c);
            }

            /**
             * Get predicate asserting free var is a char
             */
            z3::expr get_is_char(int c) {
                return get_is_char(context, c);
            }

            /**
             * Get predicate asserting free in range
             */
            z3::expr get_in_range(unsigned char lo, unsigned char hi) {
                return context.int_val((int) lo) <= get_free_variable() &&
                    get_free_variable() <= context.int_val((int) hi);
            }

            /**
             * Convert predicate to match automata transitions
             *
             * Assumed predicate specifies a character and has one free
             * variable.
             */
            z3::expr translate_predicate(z3::expr const& pred);

            /**
             * Read a char from pos of string and put in out
             *
             * Returns next position. Mostly just takes the character, but
             * \u{...} needs special handling.
             */
            static unsigned chomp_char(
                std::string const &s,
                unsigned pos,
                int &out
            );

        private:
            symbolic_automaton* re2aut(z3::expr const &re);
            symbolic_automaton* seq2aut(z3::expr const &e);

            /**
             * Converts a 1-char string to a char
             *
             * Throws an exception if not a 1-char string.
             */
            int get_char(z3::expr const &c);

            /**
             * Extract all replacements to make aut compat
             *
             * Throws exception if the translation won't work
             *
             * Seens ids is to prevent adding the same expr twice
             */
            void predicate_extract_replaceable_exprs(
                z3::expr const &pred,
                z3::expr_vector &old_expr,
                z3::expr_vector &new_expr,
                std::set<unsigned> &seen_ids
            );

            bool is_string_var(z3::expr e) {
                return e.is_seq() && e.is_const() && !e.is_string_value();
            }
    };
}

#endif
