
#include <iostream>
#include <sstream>
#include <stdlib.h>

#include "z3++.h"

#include "parikh_image_one.h"
#include "parikh_image_two.h"
#include "re2automaton.h"
#include "sym_parikh.h"
#include "sym_parikh_util.h"

using namespace std;
using namespace sym_parikh;
using namespace z3;

SymParikh::SymParikh(
    std::function<expr_vector (z3::context&)> get_smt_contents,
    SymParikhConfig config
) try : context(),
        solver(this->context),
        config(config),
        nnf_smt_contents(
            make_nnf_smt_contents(get_smt_contents(this->context))
        ),
        profile(
            context,
            this->nnf_smt_contents,
            ParikhProfileConfig(
                config.predicates_filename,
                config.preds_from_lits
            )
        ),
        expr_pred_vars(),
        predicates(),
        check_exprs(),
        check_result(unknown),
        check_model(nullptr),
        checked(false) {
    fill_predicates();

    expr_vector old_fmlas(context);
    expr_vector new_fmlas(context);

    get_len_substitutions(old_fmlas, new_fmlas);
    get_in_re_substitutions(old_fmlas, new_fmlas);
    get_seq_eq_substitutions(old_fmlas, new_fmlas);
    get_re_eq_substitutions(old_fmlas, new_fmlas);

    for (expr e : nnf_smt_contents) {
        check_exprs.push_back(e.substitute(old_fmlas, new_fmlas));
    }

    add_domain_checks();
} catch (z3::exception& e) {
    throw new SymParikhException(e.what());
} catch (ParikhProfileException *e) {
    throw new SymParikhException(e->what());
}

check_result SymParikh::check_parikh_sat() {
    if (!checked) {
        solver.push();
        for (expr e : check_exprs) {
            solver.add(e);
        }
        // cout << solver << endl;
        check_result = solver.check();
        if (check_result == sat && config.get_model)
            get_model_from_sat_solver();
        solver.pop();
    }
    return check_result;
}

void SymParikh::fill_predicates() {
    int limit = config.predicate_num_limit;
    set<unsigned> seen;

    for (expr p : profile.get_predicates()) {
        expr simp = p.simplify();
        if (seen.find(simp.id()) == seen.end()) {
            if (include_predicate(simp))
                predicates.push_back(simp);
            seen.insert(simp.id());
        }

        if (limit >= 0 && seen.size() >= limit)
            break;
    }

    if (!has_predicate(predicates, anychar()))
        predicates.push_back(anychar());
}

void SymParikh::get_len_substitutions(
    expr_vector &old_fmlas,
    expr_vector &new_fmlas
) {
    for (expr len_expr : profile.get_seq_lens()) {
        old_fmlas.push_back(len_expr);
        new_fmlas.push_back(convert_len_expr(len_expr.args()[0]));
    }
}

expr SymParikh::convert_len_expr(expr len) {
    if (len.is_seq() && len.is_const()) {
        // either variable or literal
        if (len.is_string_value()) {
            return context.int_val(len.get_string().length());
        } else {
            return getadd_string_pred_var(len, anychar());
        }
    } else {
        switch (len.decl().decl_kind()) {
        case Z3_OP_SEQ_CONCAT: {
            expr sum = context.int_val(0);
            for (expr arg : len.args())
                sum = sum + convert_len_expr(arg);
            return sum;
        }
        default:
            ostringstream ss;
            ss << "Unsupported str.len LHS " << len;
            throw new SymParikhException(ss.str());
        }
    }
}

void SymParikh::get_in_re_substitutions(
    expr_vector &old_fmlas,
    expr_vector &new_fmlas
) {
    for (expr in_re : profile.get_in_re_checks()) {
        old_fmlas.push_back(in_re);
        new_fmlas.push_back(convert_in_re_expr(in_re));
    }
    for (expr in_re : profile.get_not_in_re_checks()) {
        old_fmlas.push_back(in_re);
        new_fmlas.push_back(convert_not_in_re_expr(in_re));
    }
}

expr SymParikh::convert_in_re_expr(expr const &in_re) {
    auto args = in_re.args();
    if (args.size() < 2) {
        ostringstream ss;
        ss << "Not enough args to in_re expression " << in_re;
        throw new SymParikhException(ss.str());
    }

    vector<expr> lhs_preds;
    expr lhs = args[0];
    for (expr pred : predicates) {
        lhs_preds.push_back(convert_string_expr(lhs, pred));
    }

    expr rhs = args[1];
    shared_ptr<ParikhImage> pi(config.pi_provider->get_parikh_image(
        context,
        profile.get_positive_automaton(rhs),
        predicates,
        lhs_preds
    ));

    return pi->get_image_formula();
}

expr SymParikh::convert_not_in_re_expr(expr const &not_in_re) {
    if (not_in_re.args().size() != 1) {
        ostringstream ss;
        ss << "Malformed negation has too many arguments: " << not_in_re;
        throw new SymParikhException(ss.str());
    }

    auto args = not_in_re.args()[0].args();
    if (args.size() < 2) {
        ostringstream ss;
        ss << "Not enough args to in_re expression " << not_in_re;
        throw new SymParikhException(ss.str());
    }

    vector<expr> lhs_preds;
    expr lhs = args[0];
    for (expr pred : predicates) {
        lhs_preds.push_back(convert_string_expr(lhs, pred));
    }

    expr rhs = args[1];
    shared_ptr<ParikhImage> pi(config.pi_provider->get_parikh_image(
        context,
        profile.get_negative_automaton(rhs),
        predicates,
        lhs_preds
    ));

    return pi->get_image_formula();
}

expr SymParikh::convert_string_expr(expr const &expression, expr const &pred) {
    if (expression.is_seq() && expression.is_const()) {
        // either variable or literal
        if (expression.is_string_value()) {
            return context.int_val(
                sum_satisfying_chars(expression.get_string(), pred)
            );
        } else {
            return getadd_string_pred_var(expression, pred);
        }
    } else {
        // TODO: assert pred in predicates?
        optional<expr> expr_var = get_string_expr_pred_var(expression, pred);
        if (expr_var.has_value()) {
            return expr_var.value();
        } else {
            convert_string_expr_full(expression);
            // should not fail now!
            return get_string_expr_pred_var(expression, pred).value();
        }
    }
}

void SymParikh::convert_string_expr_full(expr const &expression) {
    switch (expression.decl().decl_kind()) {
    // str.++: expression_pred = arg1_pred + ... + argn_pred
    case Z3_OP_SEQ_CONCAT: {
        for (expr const &pred : predicates) {
            expr ep_var = getadd_string_expr_pred_var(expression, pred);
            expr sum = context.int_val(0);
            for (expr arg : expression.args())
                sum = sum + convert_string_expr(arg, pred);
            check_exprs.push_back(ep_var == sum);
        }
        break;
    }
    // str.substr x m n: take any coherent sub-image of x
    case Z3_OP_SEQ_EXTRACT: {
        if (expression.args().size() < 1) {
            ostringstream ss;
            ss << "Not enough arguments to " << expression;
            throw new SymParikhException(ss.str());
        }
        expr arg = expression.args()[0];
        for (expr const &pred : predicates) {
            expr ep_var = getadd_string_expr_pred_var(expression, pred);
            check_exprs.push_back(ep_var <= convert_string_expr(arg, pred));
        }
        add_check_is_string(expression);
        break;
    }
    case Z3_OP_SEQ_REPLACE: {
         if (expression.args().size() < 3) {
            ostringstream ss;
            ss << "Not enough arguments to " << expression;
            throw new SymParikhException(ss.str());
        }
        expr orig = expression.args()[0];
        expr to_replace = expression.args()[1];
        expr replace_by = expression.args()[2];

        // case 1: no replace occurs
        expr no_replace = context.bool_val(true);
        for (expr pred : predicates) {
            expr orig_var = convert_string_expr(orig, pred);
            expr ep_var = getadd_string_expr_pred_var(expression, pred);
            no_replace = no_replace && (orig_var == ep_var);
        }

        // case 2: a replace occurs
        expr a_replace = context.bool_val(true);
        for (expr pred : predicates) {
            expr orig_var = convert_string_expr(orig, pred);
            expr to_replace_var = convert_string_expr(to_replace, pred);
            expr replace_by_var = convert_string_expr(replace_by, pred);
            expr ep_var = getadd_string_expr_pred_var(expression, pred);
            a_replace = a_replace && (
                ep_var == orig_var - to_replace_var + replace_by_var
            );
        }

        // case 3: to_replace is empty, append replace_by (weird SMTLIB
        // semantics)
        expr e_replace = context.bool_val(true);
        for (expr pred : predicates) {
            expr orig_var = convert_string_expr(orig, pred);
            expr to_replace_var = convert_string_expr(to_replace, pred);
            expr replace_by_var = convert_string_expr(replace_by, pred);
            expr ep_var = getadd_string_expr_pred_var(expression, pred);
            e_replace = e_replace && (
                ep_var == orig_var + replace_by_var && to_replace_var == 0
            );
        }

        check_exprs.push_back(no_replace || a_replace || e_replace);
        add_check_is_string(expression);
        break;
    }
    default:
        ostringstream ss;
        ss << "Unsupported string expression " << expression;
        throw new SymParikhException(ss.str());
    }
}

unsigned SymParikh::sum_satisfying_chars(string const &literal, expr const &pred) {
    expr fv = context.string_const(re2automaton::free_variable.c_str());

    unsigned count = 0;

    unsigned pos = 0;
    while (pos < literal.length()) {
        int c;
        pos = re2automaton::chomp_char(literal, pos, c);

        solver.push();
        expr f = re2automaton::get_is_char(context, c) && pred;
        solver.add(f);
        if (solver.check() == sat)
            count += 1;
        solver.pop();
    }

    return count;
}

expr SymParikh::getadd_string_expr_pred_var(
    string const &prefix,
    expr const &expression,
    expr const &predicate
) {
    auto key = make_pair(expression.id(), predicate.id());
    auto value = expr_pred_vars.find(key);
    if (value == expr_pred_vars.end()) {
        string name = prefix
            + to_string(expression.id())
            + "_"
            + to_string(predicate.id());
        expr vble = context.int_const(name.c_str());
        expr_pred_vars.insert({key, vble});
        return vble;
    } else {
        return value->second;
    }
}

expr SymParikh::getadd_string_expr_pred_var(
    expr const &expression,
    expr const &predicate
) {
    return getadd_string_expr_pred_var("se_", expression, predicate);
}

expr SymParikh::getadd_string_pred_var(
    expr const &string_var,
    expr const &predicate
) {
    return getadd_string_expr_pred_var("sc_", string_var, predicate);
}

optional<expr> SymParikh::get_string_expr_pred_var(
    expr const &expression,
    expr const &predicate
) const {
    auto key = make_pair(expression.id(), predicate.id());
    auto value = expr_pred_vars.find(key);
    if (value == expr_pred_vars.end())
        return nullopt;
    else
        return make_optional(value->second);
}

optional<expr> SymParikh::get_string_pred_var(
    expr const &string_var,
    expr const &predicate
) const {
    return get_string_expr_pred_var(string_var, predicate);
}

bool SymParikh::has_predicate(
    std::vector<z3::expr> const& predicates,
    z3::expr predicate
) const {
    for (int i = 0; i < predicates.size(); i++) {
        if (predicates[i].id() == predicate.id())
            return true;
    }
    return false;
}

expr_vector SymParikh::make_nnf_smt_contents(expr_vector const &contents) {
    goal g(context);
    for (expr e : contents)
        g.add(e);
    tactic nnf(context, "nnf");
    apply_result r = nnf.apply(g);

    if (r.size() != 1) {
        ostringstream ss;
        ss << "Only expected one goal from simplified formula. Got " << r;
        throw new SymParikhException(ss.str());
    }

    expr_vector result(context);
    for (int i = 0; i < r[0].size(); i++) {
        result.push_back(r[0][i]);
    }

    return result;
}

vector<expr> SymParikh::get_string_vars() const {
    vector<expr> vars;
    set<unsigned> ids;
    for (expr const &smt : nnf_smt_contents)
        get_string_vars(smt, ids, vars);
    return vars;
}

void SymParikh::get_string_vars(
    expr const &e,
    set<unsigned> &ids,
    vector<expr> &vars
) const {
    if (e.is_seq() && e.is_const() && !e.is_string_value()) {
        if (ids.find(e.id()) == ids.end()) {
            vars.push_back(e);
            ids.insert(e.id());
        }
    } else {
        for (expr const &arg : e.args())
            get_string_vars(arg, ids, vars);
    }
}

bool SymParikh::include_predicate(expr const &pred) {
    switch (pred.decl().decl_kind()) {
    case Z3_OP_SEQ_IN_RE:
    case Z3_OP_EQ:
    case Z3_OP_TRUE:
        return true;
    case Z3_OP_AND:
        // rough way of checking lo <= x && x <= hi
        switch (pred.args().size()) {
        case 0:
            return false;
        case 1:
            return include_predicate(pred.args()[0]);
        case 2:
            // only keep (lo <= x) and (x <= hi)
            return (
                pred.args()[0].decl().decl_kind() == Z3_OP_LE ||
                pred.args()[0].decl().decl_kind() == Z3_OP_GE
            ) && (
                pred.args()[1].decl().decl_kind() == Z3_OP_LE ||
                pred.args()[1].decl().decl_kind() == Z3_OP_GE
            );
        default:
            return false;
        }
    default:
        return false;
    }
}

void SymParikh::add_domain_checks() {
    // all var_pred counts are positive
    for (auto kv : expr_pred_vars) {
        check_exprs.push_back(kv.second >= 0);
    }

    // and all var_pred for a var are the Parikh image of an actual
    // string
    for (expr const &var : get_string_vars()) {
        add_check_is_string(var);
    }
}

void SymParikh::add_check_is_string(expr const &str_expr) {
    vector<expr> pred_vars;
    for (expr const &pred : predicates)
        pred_vars.push_back(getadd_string_expr_pred_var(str_expr, pred));
    shared_ptr<ParikhImage> pi(config.pi_provider->get_parikh_image(
        context,
        re_full(context.string_sort()),
        predicates,
        pred_vars
    ));
    check_exprs.push_back(pi->get_image_formula());
}

void SymParikh::get_seq_eq_substitutions(
    expr_vector &old_fmlas,
    expr_vector &new_fmlas
) {
    for (expr const &eq : profile.get_seq_equalities()) {
        expr lhs = eq.args()[0];
        expr rhs = eq.args()[1];

        expr pred_conditions = context.bool_val(true);

        for (expr const &pred : predicates) {
            expr plhs = convert_string_expr(lhs, pred);
            expr prhs = convert_string_expr(rhs, pred);

            expr condition = context.bool_val(true);

            switch (eq.decl().decl_kind()) {
            case Z3_OP_EQ:
                condition = (plhs == prhs);
                break;
            case Z3_OP_SEQ_PREFIX:
            case Z3_OP_SEQ_SUFFIX:
                condition = (plhs <= prhs);
                break;
            case Z3_OP_SEQ_CONTAINS:
                condition = (plhs >= prhs);
                break;
            default: {
                ostringstream ss;
                ss << "Unexpected seq eq op " << eq;
                throw new SymParikhException(ss.str());
            }}

            pred_conditions = pred_conditions && condition;
        }

        old_fmlas.push_back(eq);
        new_fmlas.push_back(pred_conditions);
    }
}

void SymParikh::get_re_eq_substitutions(
    expr_vector &old_fmlas,
    expr_vector &new_fmlas
) {
    // re1 == re2 becomes image(re1) == image(re2)
    // I.e. forall x . image(re1)(x) == image(re2)(x)
    // This is overapprox since all equal regex have same Parikh image,
    // but some unequal ones do too.
    for (expr const &eq : profile.get_re_equalities()) {
        expr lhs = eq.args()[0];
        expr rhs = eq.args()[1];

        // annoying there's a type mismatch so we need to vectors
        // TODO: refactor to expr_vector everywhere
        expr_vector pred_vars(context);
        std::vector<expr> std_pred_vars;

        for (expr const &pred : predicates) {
            string name = "c_" + to_string(pred.id());
            expr var = context.int_const(name.c_str());
            pred_vars.push_back(var);
            std_pred_vars.push_back(var);
        }

        expr pi_lhs = config.pi_provider->get_parikh_image(
                context,
                profile.get_positive_automaton(lhs),
                predicates,
                std_pred_vars
            )->get_image_formula();

        expr pi_rhs = config.pi_provider->get_parikh_image(
                context,
                profile.get_positive_automaton(rhs),
                predicates,
                std_pred_vars
            )->get_image_formula();

        old_fmlas.push_back(eq);
        new_fmlas.push_back(forall(pred_vars, pi_lhs == pi_rhs));
    }

    // re1 != re2 becomes
    //  exists x .
    //      image(re1)(x) and image(comp(re2))(x)
    //      or
    //      image(comp(re1))(x) and image(re2)(x)
    // I.e. is some element of Parikh image accepted by one and the
    // complement of the other
    // This is overapprox because the image overapproxes both re and
    // comp(re).
    for (expr const &neq : profile.get_re_notequalities()) {
        expr lhs = neq.args()[0].args()[0];
        expr rhs = neq.args()[0].args()[1];

        // annoying there's a type mismatch so we need to vectors
        // TODO: refactor to expr_vector everywhere
        expr_vector pred_vars(context);
        std::vector<expr> std_pred_vars;

        for (expr const &pred : predicates) {
            string name = "c_" + to_string(pred.id());
            expr var = context.int_const(name.c_str());
            pred_vars.push_back(var);
            std_pred_vars.push_back(var);
        }

        expr pi_lhs_pos = config.pi_provider->get_parikh_image(
                context,
                profile.get_positive_automaton(lhs),
                predicates,
                std_pred_vars
            )->get_image_formula();

        expr pi_lhs_neg = config.pi_provider->get_parikh_image(
                context,
                profile.get_negative_automaton(lhs),
                predicates,
                std_pred_vars
            )->get_image_formula();

        expr pi_rhs_pos = config.pi_provider->get_parikh_image(
                context,
                profile.get_positive_automaton(rhs),
                predicates,
                std_pred_vars
            )->get_image_formula();

        expr pi_rhs_neg = config.pi_provider->get_parikh_image(
                context,
                profile.get_negative_automaton(rhs),
                predicates,
                std_pred_vars
            )->get_image_formula();

        old_fmlas.push_back(neq);
        new_fmlas.push_back(exists(pred_vars,
            (pi_lhs_pos && pi_rhs_neg) || (pi_lhs_neg && pi_rhs_pos)
        ));
    }
}

void SymParikh::get_model_from_sat_solver() {
    model const m = solver.get_model();
    check_model = std::shared_ptr<SymParikhModel>(new SymParikhModel());

    vector<expr> vars = get_string_vars();
    for (expr const &v : vars) {
        for (expr const &p : predicates) {
            optional<expr> const &var = get_string_pred_var(v, p);
            if (var.has_value()) {
                check_model->set_count(v, p, m.eval(var.value()).get_numeral_uint());
            }
        }
    }
}

void SymParikhModel::set_count(
    expr const &string_var,
    expr const &pred,
    unsigned count
) {
    if (counts.find(string_var.id()) == counts.end()) {
        counts.insert({ string_var.id(), std::map<unsigned, unsigned>() });
        string_vars.push_back(string_var);
    }
    counts.find(string_var.id())->second.insert_or_assign(pred.id(), count);

    if (seen_predicates.find(pred.id()) == seen_predicates.end()) {
        seen_predicates.insert(pred.id());
        predicates.push_back(pred);
    }
}

unsigned SymParikhModel::get_count(expr const &string_var, expr const &pred) {
    auto found_string_map = counts.find(string_var.id());
    if (found_string_map == counts.end())
        return 0;
    auto string_map = found_string_map->second;

    auto found_pred_val = string_map.find(pred.id());
    if (found_pred_val == string_map.end())
        return 0;

    return found_pred_val ->second;
}

