#ifndef SYM_PARIKH_H
#define SYM_PARIKH_H

#include <functional>
#include <iostream>
#include <map>
#include <optional>
#include <string>

#include "z3++.h"

#include "parikh_image_provider.h"
#include "parikh_profile.h"

namespace sym_parikh {
    class SymParikhException : public std::exception {
        public:
            SymParikhException(std::string const &reason)
                    : reason(reason) { }

            virtual const char* what() const throw() {
                return reason.c_str();
            }

        private:
            std::string const reason;
    };

    struct SymParikhConfig {
        std::shared_ptr<ParikhImageProvider> pi_provider;
        std::shared_ptr<std::string> predicates_filename;
        bool preds_from_lits;
        bool get_model;
        int predicate_num_limit;

        SymParikhConfig() : SymParikhConfig(
            std::shared_ptr<ParikhImageProvider>(
                new ParikhImageProviderOne()
            ),
            nullptr,
            false,
            false,
            -1
        ) { }

        /**
         * Build a config
         *
         * @param pi_provider a Parikh image provider @param
         * predicates_filename a file to read predicates from or null if to be
         * infered.
         * @param preds_from_lits whether to infer predicates from chars in
         * string literals
         * @param predicate_num_limit only take the first n predicates.
         * -1 means take all.
         */
        SymParikhConfig(
            std::shared_ptr<ParikhImageProvider> pi_provider,
            std::shared_ptr<std::string> predicates_filename,
            bool preds_from_lits,
            bool get_model,
            int predicate_num_limit
        ) : pi_provider(pi_provider),
            predicates_filename(predicates_filename),
            preds_from_lits(preds_from_lits),
            get_model(get_model),
            predicate_num_limit(predicate_num_limit) { }
    };

    class SymParikhModel {
        public:
            SymParikhModel() :
                counts(),
                string_vars(),
                predicates(),
                seen_predicates() { }

            /**
             * Set the count of chars in the string_var statisfying pred
             */
            void set_count(
                z3::expr const &string_var,
                z3::expr const &pred,
                unsigned count
            );

            /**
             * Get the count of chars in the string_var statisfying pred
             */
            unsigned get_count(
                z3::expr const &string_var,
                z3::expr const &pred
            );

            /**
             * Get the predicates for each string in the model
             */
            std::vector<z3::expr> const &get_predicates() {
                return predicates;
            }

            /**
             * Get the string vars in the model
             */
            std::vector<z3::expr> const &get_string_vars() {
                return string_vars;
            }

        private:
            /**
             * Map from original string var id to a map from predicate id to
             * predicate counts.
             */
            std::map<unsigned, std::map<unsigned, unsigned>> counts;
            std::vector<z3::expr> string_vars;
            std::vector<z3::expr> predicates;
            std::set<unsigned> seen_predicates;
    };

    class SymParikh {
        public:
            static SymParikh from_file(std::string const &smt_filename) {
                return SymParikh(
                    [smt_filename](z3::context &c) {
                        return c.parse_file(smt_filename.c_str());
                    },
                    SymParikhConfig()
                );
            }

            static SymParikh from_string(std::string const &smt_string) {
                return SymParikh(
                    [smt_string](z3::context &c) {
                        return c.parse_string(smt_string.c_str());
                    },
                    SymParikhConfig()
                );
            }

            static SymParikh from_file(
                std::string const &smt_filename,
                SymParikhConfig config
            ) {
                return SymParikh(
                    [smt_filename](z3::context &c) {
                        return c.parse_file(smt_filename.c_str());
                    },
                    config
                );
            }

            static SymParikh from_string(
                std::string const &smt_string,
                SymParikhConfig config
            ) {
                return SymParikh(
                    [smt_string](z3::context &c) {
                        return c.parse_string(smt_string.c_str());
                    },
                    config
                );
            }

            /**
             * Check if the Parikh version is sat
             *
             * Runs tests the first time, returns stored result after.
             */
            z3::check_result check_parikh_sat();

            /**
             * Write the model to ostream
             *
             * Assumes check_parikh_sat called and got a sat result. Null if no
             * model.
             */
            std::shared_ptr<SymParikhModel> get_model() const {
                return check_model;
            }

            std::vector<z3::expr> get_predicates() const { return predicates; }

        protected:
            /**
             * Do as-yet unspecified processing on the SMT file
             *
             * This function likely to be refined when the purpose of the
             * tool becomes clearer. For now, just used to test parsing a
             * regular expression and doing the translations.
             *
             * @param get_smt_contents a function to get the smt
             * contents from a given context (either by parse_file or
             * parse_string, see named constructors above).
             * @param parikh_image_provider a provider of Parikh Images
             * using one of the different methods we have
             * @param predicates_filename an SMT file containing a
             * sequence of assertions over a free variable x to use as
             * predicates
             */
            SymParikh(
                std::function<z3::expr_vector (z3::context&)> get_smt_contents,
                SymParikhConfig config
            );

        private:
            /**
             * The z3 context and solver we'll use throughout
             */
            z3::context context;
            z3::solver solver;
            SymParikhConfig config;

            /**
             * The parsed and nnf SMT file/string
             */
            z3::expr_vector nnf_smt_contents;

            /**
             * Data about the SMT being processed
             */
            ParikhProfile profile;

            /**
             * The set of predicates our analysis should take care of
             */
            std::vector<z3::expr> predicates;

            /**
             * Count variables for (string_var, predicate).
             *
             * I.e. how many times predicate is true in string_var. Map
             * is from expression ids, (id, id) -> expr.
             */
            std::map<std::pair<unsigned, unsigned>, z3::expr> expr_pred_vars;

            /**
             * The final expressions for checking unsat
             *
             * I.e. a Parikh image version of the input SMT file. Stores
             * check results for later use.
             */
            std::vector<z3::expr> check_exprs;
            z3::check_result check_result;
            std::shared_ptr<SymParikhModel> check_model;
            bool checked;

            /**
             * Fill predicates array with preds needed for analysis
             */
            void fill_predicates();

            /**
             * Get list of len formula substitutions
             *
             * E.g. (str.len x) becomes #anychar in x
             *
             * @param old_fmlas array to fill with the parts of the
             * formula to be replaced.
             * @param new_fmlas array to fill with the replacements of
             * the old formulas.
             */
            void get_len_substitutions(
                z3::expr_vector &old_fmlas,
                z3::expr_vector &new_fmlas
            );

            /**
             * Get list of in_re formula substitutions
             *
             * E.g. (str.in_re x r) becomes a formula over the predicate
             * variables for the string variable. E.g. f(x_p1, x_p2) if
             * there are two predicates and f is the Parikh image of r
             * against p1 and p2.
             *
             * @param old_fmlas array to fill with the parts of the
             * formula to be replaced.
             * @param new_fmlas array to fill with the replacements of
             * the old formulas.
             */
            void get_in_re_substitutions(
                z3::expr_vector &old_fmlas,
                z3::expr_vector &new_fmlas
            );

            /**
             * Translate the argument of a len expression to Parikh
             *
             * E.g. x + y becomes #x_anychar + #y_anychar
             */
            z3::expr convert_len_expr(z3::expr len);

            /**
             * Translate an in_re expression to Parikh
             *
             * E.g. (str.in_re x re) becomes f(x_p1, x_p2) where f is
             * the Parikh image of re against predicates p1, p2. The
             * predicates come from the profile.
             */
            z3::expr convert_in_re_expr(z3::expr const &in_re);

            /**
             * Translate a not in_re expression to Parikh
             *
             * E.g. (not (str.in_re x re)) becomes f'(x_p1, x_p2) where
             * f' is the Parikh image of the negation of re against
             * predicates p1, p2. The predicates come from the profile.
             */
            z3::expr convert_not_in_re_expr(z3::expr const &not_in_re);

            /**
             * Convert a string value to Parikh image against pred
             *
             * Returns an expression that is E.g. x ++ y becomes x_pred +
             * y_pred. This may trigger the addition of new variables and side
             * conditions to check_exprs if e.g. the expression is a
             * str.replace and all predicates need to be handled at once,
             * perhaps with the addition of new variables.
             *
             * Requires pred in predicates
             */
            z3::expr convert_string_expr(
                z3::expr const &expression,
                z3::expr const &pred
            );

            /**
             * Do a full conversion of expression for all predicates
             *
             * Adds conditions to check_exprs if needed. E.g. x + y will
             * introduce new variables expr_pred for each pred, that is the sum
             * x_pred + y_pred. This sum will be added to check_exprs and the
             * var expr_pred will be available via get_string_expr_pred_var.
             */
            void convert_string_expr_full(
                z3::expr const &expression
            );

            unsigned sum_satisfying_chars(
                std::string const &literal,
                z3::expr const &pred
            );

            /**
             * Number of times pred is true for given string expr
             *
             * With custom name prefix
             */
            z3::expr getadd_string_expr_pred_var(
                std::string const &prefix,
                z3::expr const &expr,
                z3::expr const &predicate
            );

            /**
             * Number of times pred is true for given string expr
             */
            z3::expr getadd_string_expr_pred_var(
                z3::expr const &expr,
                z3::expr const &predicate
            );

            /**
             * Number of times pred is true for string var
             */
            z3::expr getadd_string_pred_var(
                z3::expr const &string_var,
                z3::expr const &predicate
            );

            /**
             * Get var for how many times expr matches pred
             */
            std::optional<z3::expr> get_string_expr_pred_var(
                z3::expr const &expression,
                z3::expr const &predicate
            ) const;

            /**
             * Get var for how many times string matches pred
             */
            std::optional<z3::expr> get_string_pred_var(
                z3::expr const &string_var,
                z3::expr const &predicate
            ) const;

            /**
             * Check if the list of predicates has predicate
             */
            bool has_predicate(
                std::vector<z3::expr> const& predicates,
                z3::expr predicate
            ) const;

            z3::expr anychar() { return context.bool_val(true); }

            /**
             * Converts contents into negation normal form
             */
            z3::expr_vector make_nnf_smt_contents(
                z3::expr_vector const &contents
            );

            /**
             * Get all string vars from smt
             */
            std::vector<z3::expr> get_string_vars() const;

            /**
             * Get all string vars from smt expr
             *
             * Stores in vars array. ids is a set of ids of variables
             * already in vars. Also added to.
             */
            void get_string_vars(
                z3::expr const &e,
                std::set<unsigned> &ids,
                std::vector<z3::expr> &vars
            ) const;

            /**
             * Whether to include this predicate as part of Parikh image
             */
            bool include_predicate(z3::expr const &pred);

            /**
             * Assert variables are in the right domain
             *
             * E.g. string predicate counts are >= 0
             */
            void add_domain_checks();

            /**
             * Assert pred_vars for str_expr is a string
             *
             * That is pred_vars for str_expr can be the Parikh Image of .*.
             * Assumes getadd_string_expr_pred_var returns variables for
             * str_expr.
             *
             * Any sequence of numbers might not add up. E.g. if the predicates
             * are =a and true, then there should always be at least as many
             * true as =a.
             */
            void add_check_is_string(z3::expr const &str_expr);

            /**
             * Get (= string1 string2) replacements and neg
             *
             * @param old_fmlas array to fill with the parts of the
             * formula to be replaced.
             * @param new_fmlas array to fill with the replacements of
             * the old formulas.
             */
            void get_seq_eq_substitutions(
                z3::expr_vector &old_fmlas,
                z3::expr_vector &new_fmlas
            );

            /**
             * Get (= re1 re2) replacements and neg
             *
             * @param old_fmlas array to fill with the parts of the
             * formula to be replaced.
             * @param new_fmlas array to fill with the replacements of
             * the old formulas.
             */
            void get_re_eq_substitutions(
                z3::expr_vector &old_fmlas,
                z3::expr_vector &new_fmlas
            );

            /**
             * Assuming the solver has just done a check that is sat, get model
             *
             * Stores in check_model
             */
            void get_model_from_sat_solver();
    };
}

#endif
