
#ifndef SYMBOLIC_AUTOMATON_H
#define SYMBOLIC_AUTOMATON_H

#include <algorithm>
#include <iostream>
#include <map>
#include <optional>
#include <set>
#include <vector>

#include "z3++.h"

namespace sym_parikh {
    /**
     * Code very much taken from z3 internal automaton class
     */
    class symbolic_automaton {
    public:
        class move {
            std::optional<z3::expr> m_t;
            unsigned m_src;
            unsigned m_dst;

        public:
            move(
                unsigned s,
                unsigned d,
                std::optional<z3::expr const> const &t
            ) : m_t(t), m_src(s), m_dst(d) { }

            move(unsigned s, unsigned d, z3::expr const &t)
                    : m_t(std::make_optional(t)), m_src(s), m_dst(d) { }

            move(unsigned s, unsigned d)
                    : m_t(std::nullopt), m_src(s), m_dst(d) { }

            move(move const& other)
                    : m_t(other.m_t), m_src(other.m_src), m_dst(other.m_dst) { }

            move(move &&other) noexcept
                    : m_t(std::nullopt),
                        m_src(other.m_src),
                        m_dst(other.m_dst) {
                std::swap(m_t, other.m_t);
            }

            move& operator=(move const &other) {
                m_t = other.m_t;
                m_src = other.m_src;
                m_dst = other.m_dst;
                return *this;
            }

            unsigned dst() const { return m_dst; }
            unsigned src() const { return m_src; }

            /**
             * std::nullopt if epsilon
             */
            std::optional<z3::expr const> t() const { return m_t; }
            bool is_epsilon() const { return !m_t.has_value(); }
        };
        typedef std::set<move> moves;

        // to permit use of maps with move keys
        friend bool operator<(move const &m1, move const &m2) {
            if (m1.src() != m2.src())
                return m1.src() < m2.src();
            if (m1.dst() != m2.dst())
                return m1.dst() < m2.dst();
            return (move_t_lt(m1.t(), m2.t()));
        }
    private:
        std::vector<moves> m_delta;
        std::vector<moves> m_delta_inv;
        unsigned m_init;
        std::set<unsigned> m_final_set;
        std::vector<unsigned> m_final_states;

        // local data-structures
        mutable std::set<unsigned> m_visited;
        mutable std::vector<unsigned> m_todo;

        struct default_display {
            std::ostream& display(std::ostream& out, z3::expr const &t) {
                return out << t;
            }
        };

    public:

        // The empty automaton:
        symbolic_automaton() : m_init(0) {
            m_delta.push_back(moves());
            m_delta_inv.push_back(moves());
        }

        // create an automaton from initial state, final states, and moves
        symbolic_automaton(
            unsigned init,
            std::vector<unsigned> &final,
            moves const &mvs
        ) {
            m_init = init;
            m_delta.push_back(moves());
            m_delta_inv.push_back(moves());
            for (unsigned f : final) {
                add_to_final_states(f);
            }
            for (move const& mv : mvs) {
                unsigned n = std::max(mv.src(), mv.dst());
                if (n >= m_delta.size()) {
                    m_delta.resize(n+1, moves());
                    m_delta_inv.resize(n+1, moves());
                }
                add(mv);
            }
        }

        // create an automaton that accepts a sequence.
        symbolic_automaton(std::vector<z3::expr> const &seq): m_init(0) {
            m_delta.resize(seq.size()+1, moves());
            m_delta_inv.resize(seq.size()+1, moves());
            for (unsigned i = 0; i < seq.size(); ++i) {
                m_delta[i].insert(move(i, i + 1, seq[i]));
                m_delta[i + 1].insert(move(i, i + 1, seq[i]));
            }
            add_to_final_states(seq.size());
        }

        // The automaton that accepts t
        symbolic_automaton(z3::expr const &t) : m_init(0) {
            m_delta.resize(2, moves());
            m_delta_inv.resize(2, moves());
            add_to_final_states(1);
            add(move(0, 1, t));
        }

        symbolic_automaton(symbolic_automaton const &other)
            : m_delta(other.m_delta),
            m_delta_inv(other.m_delta_inv),
            m_init(other.m_init),
            m_final_set(other.m_final_set),
            m_final_states(other.m_final_states) { }

        // create the automaton that accepts the empty string/sequence only.
        static symbolic_automaton* mk_epsilon() {
            moves mvs;
            std::vector<unsigned> final;
            final.push_back(0);
            return new symbolic_automaton(0, final, mvs);
        }

        // create the automaton that accepts nothing
        static symbolic_automaton* mk_none() {
            moves mvs;
            std::vector<unsigned> final;
            return new symbolic_automaton(0, final, mvs);
        }

        // create the automaton with a single state on condition t.
        static symbolic_automaton* mk_loop(z3::expr &t) {
            moves mvs;
            std::vector<unsigned> final;
            final.push_back(0);
            mvs.insert(move(0, 0, t));
            return new symbolic_automaton(0, final, mvs);
        }

        static symbolic_automaton* clone(symbolic_automaton const &a) {
            moves mvs;
            std::vector<unsigned> final;
            append_moves(0, a, mvs);
            append_final(0, a, final);
            return new symbolic_automaton(a.init(), final, mvs);
        }

        symbolic_automaton* clone() const {
            return clone(*this);
        }

        // create the sum of disjoint automata
        static symbolic_automaton* mk_union(
            symbolic_automaton const &a,
            symbolic_automaton const &b
        ) {
            if (a.is_empty()) {
                return b.clone();
            }
            if (b.is_empty()) {
                return a.clone();
            }
            moves mvs;
            std::vector<unsigned> final;
            unsigned offset1 = 1;
            unsigned offset2 = a.num_states() + 1;
            mvs.insert(move(0, a.init() + offset1));
            mvs.insert(move(0, b.init() + offset2));
            append_moves(offset1, a, mvs);
            append_moves(offset2, b, mvs);
            append_final(offset1, a, final);
            append_final(offset2, b, final);
            return new symbolic_automaton(0, final, mvs);
        }

        static symbolic_automaton* mk_opt(symbolic_automaton const &a) {
            moves mvs;
            std::vector<unsigned> final;
            unsigned offset = 0;
            unsigned init = a.init();
            if (!a.initial_state_is_source()) {
                offset = 1;
                init = 0;
                mvs.insert(move(0, a.init() + offset));
            }
            if (a.is_empty()) {
                return a.clone();
            }

            mvs.insert(move(init, a.final_state() + offset));
            append_moves(offset, a, mvs);
            append_final(offset, a, final);
            return new symbolic_automaton(init, final, mvs);
        }

        // concatenate accepting languages
        static symbolic_automaton* mk_concat(
            symbolic_automaton const &a,
            symbolic_automaton const &b
        ) {
            if (a.is_empty()) {
                return a.clone();
            }
            if (b.is_empty()) {
                return b.clone();
            }
            if (a.is_epsilon()) {
                return b.clone();
            }
            if (b.is_epsilon()) {
                return a.clone();
            }

            moves mvs;
            std::vector<unsigned> final;
            unsigned init = 0;
            unsigned offset1 = 1;
            unsigned offset2 = a.num_states() + offset1;
            mvs.insert(move(0, a.init() + offset1));
            append_moves(offset1, a, mvs);
            for (unsigned i = 0; i < a.m_final_states.size(); ++i) {
                mvs.insert(move(a.m_final_states[i] + offset1, b.init() + offset2));
            }
            append_moves(offset2, b, mvs);
            append_final(offset2, b, final);

            return new symbolic_automaton(init, final, mvs);
        }

        static symbolic_automaton* mk_reverse(symbolic_automaton const &a) {
            if (a.is_empty()) {
                return new symbolic_automaton();
            }
            moves mvs;
            for (unsigned i = 0; i < a.m_delta.size(); ++i) {
                moves const& mvs1 = a.m_delta[i];
                for (move const &mv : mvs1) {
                    mvs.insert(move(mv.dst(), mv.src(), mv.t()));
                }
            }
            std::vector<unsigned> final;
            unsigned init;
            final.push_back(a.init());
            if (a.m_final_states.size() == 1) {
                init = a.m_final_states[0];
            }
            else {
                init = a.num_states();
                for (unsigned st : a.m_final_states) {
                    mvs.insert(move(init, st));
                }
            }
            return new symbolic_automaton(init, final, mvs);
        }

        static symbolic_automaton* mk_complement(
            z3::context &context,
            symbolic_automaton const &a
        ) {
            return mk_determinstic_param(context, a, true);
        }

        static symbolic_automaton* mk_determinstic_param(
            z3::context &context,
            symbolic_automaton const &a,
            bool flip_acceptance
        ) {
            std::vector<
                std::pair<std::map<unsigned, bool>, z3::expr>
            > min_terms;
            std::vector<z3::expr> predicates;

            std::map<std::set<unsigned>, unsigned> s2id; // set of states to unique id
            std::vector<std::set<unsigned>> id2s;        // unique id to set of b-states
            std::set<unsigned> set;
            std::vector<unsigned> vector;
            moves new_mvs;                               // moves in the resulting automaton
            std::vector<unsigned> new_final_states;      // new final states
            unsigned p_state_id = 0;                     // next state identifier

            // adds non-final states of a to final if flipping and final otherwise
            std::vector<unsigned> init_states;
            a.get_epsilon_closure(a.init(), init_states);
            for (unsigned s : init_states) {
                set.insert(s);
            }
            if (a.is_final_configuration(set) != flip_acceptance) {
                new_final_states.push_back(p_state_id);
            }

            s2id[set] = p_state_id++; // the index to the initial state is 0
            id2s.push_back(set);

            std::vector<std::set<unsigned>> todo; //States to visit
            todo.push_back(set);

            std::set<unsigned> state;
            moves mvsA;

            new_mvs.clear();

            // or just make todo a vector whose indices coincide with state_id.
            while (!todo.empty()) {
                std::set<unsigned> state = todo.back();

                unsigned state_id = s2id[state];
                todo.pop_back();
                mvsA.clear();

                min_terms.clear();
                predicates.clear();

                a.get_moves_from_states(state, mvsA);

                std::set<unsigned> seen_preds;
                for (move const &m : mvsA) {
                    // no e-trans because of e-closure above
                    if (!m.is_epsilon()) {
                        if (
                            seen_preds.find(m.t().value().id())
                                == seen_preds.end()
                        ) {
                            // needs a stable reference, not p
                            predicates.push_back(m.t().value());
                            seen_preds.insert(m.t().value().id());
                        }
                    }
                }

                min_terms = a.generate_min_terms(context, predicates);
                for (unsigned j = 0; j < min_terms.size(); ++j) {
                    set = std::set<unsigned>();
                    for (move const &m : mvsA) {
                        // no e-trans because of e-closure above
                        if (!m.is_epsilon()) {
                            unsigned pid = m.t().value().id();
                            if (min_terms[j].first[pid])
                                set.insert(m.dst());
                        }
                    }

                    bool is_new = s2id.find(set) == s2id.end();
                    if (is_new) {
                        if (a.is_final_configuration(set) != flip_acceptance) {
                            new_final_states.push_back(p_state_id);
                        }

                        s2id[set] = p_state_id++;
                        id2s.push_back(set);
                        todo.push_back(set);
                    }
                    new_mvs.insert(move(
                        state_id,
                        s2id[set],
                        min_terms[j].second
                    ));
                }
            }

            if (new_final_states.empty()) {
                return new symbolic_automaton();
            }

            return new symbolic_automaton(0, new_final_states, new_mvs);
        }

        /**
         * Product the automata
         *
         * Context needed to do some SAT checks on the way and build new
         * formula
         */
        static symbolic_automaton* mk_product(
            z3::context &context,
            symbolic_automaton const &a,
            symbolic_automaton const &b
        ) {
            z3::solver solver(context);

            std::map<std::pair<unsigned, unsigned>, unsigned> pair2id;
            std::pair<unsigned, unsigned> init_pair(a.init(), b.init());
            std::vector<std::pair<unsigned, unsigned>> todo;
            todo.push_back(init_pair);
            pair2id[init_pair] = 0;
            moves mvs;
            std::vector<unsigned> final;
            std::vector<unsigned> a_init, b_init;
            a.get_epsilon_closure(a.init(), a_init);
            bool a_init_is_final = false, b_init_is_final = false;
            for (unsigned ia : a_init) {
                if (a.is_final_state(ia)) {
                    a_init_is_final = true;
                    b.get_epsilon_closure(b.init(), b_init);
                    for (unsigned ib : b_init) {
                        if (b.is_final_state(ib)) {
                            b_init_is_final = true;
                            final.push_back(0);
                            break;
                        }
                    }
                    break;
                }
            }

            unsigned n = 1;
            moves mvsA, mvsB;
            while (!todo.empty()) {
                std::pair<unsigned, unsigned> curr_pair = todo.back();
                todo.pop_back();
                unsigned src = pair2id[curr_pair];
                mvsA.clear(); mvsB.clear();
                a.get_moves_from(curr_pair.first,  mvsA, true);
                b.get_moves_from(curr_pair.second, mvsB, true);
                for (move const &mvA : mvsA) {
                    for (move const &mvB : mvsB) {
                        // guaranteed non-epsilon from e-closure above
                        z3::expr ab =
                            mvA.t().value() && mvB.t().value();
                        solver.reset();
                        solver.add(ab);
                        auto is_sat = solver.check();
                        if (is_sat == z3::unsat) {
                            continue;
                        }
                        else if (is_sat == z3::unknown) {
                            return nullptr;
                        }
                        std::pair<unsigned, unsigned> tgt_pair(
                            mvA.dst(), mvB.dst()
                        );
                        unsigned tgt;
                        if (pair2id.find(tgt_pair) == pair2id.end()) {
                            tgt = n++;
                            pair2id[tgt_pair] = tgt;
                            todo.push_back(tgt_pair);
                            if (
                                a.is_final_state(tgt_pair.first) &&
                                b.is_final_state(tgt_pair.second)
                            ) {
                                final.push_back(tgt);
                            }
                        } else {
                            tgt = pair2id[tgt_pair];
                        }
                        mvs.insert(move(src, tgt, ab));
                    }
                }
            }

            if (final.empty()) {
                return new symbolic_automaton();
            }
            std::vector<moves> inv(n, moves());
            for (move const &mv : mvs) {
                inv[mv.dst()].insert(move(mv.dst(), mv.src(), mv.t()));
            }

            std::vector<bool> back_reachable(n, false);
            for (unsigned f : final) {
                back_reachable[f] = true;
            }

            std::vector<unsigned> stack(final);
            while (!stack.empty()) {
                unsigned state = stack.back();
                stack.pop_back();
                moves const& mvs = inv[state];
                for (move const &mv : mvs) {
                    state = mv.dst();
                    if (!back_reachable[state]) {
                        back_reachable[state] = true;
                        stack.push_back(state);
                    }
                }
            }

            moves mvs1;
            for (move const &mv : mvs) {
                if (back_reachable[mv.dst()]) {
                    mvs1.insert(mv);
                }
            }
            if (mvs1.empty()) {
                if (a_init_is_final && b_init_is_final) {
                    // special case: automaton has no moves, but the initial state is final on both sides
                    // this results in the automaton which accepts the empty sequence and nothing else
                    final.clear();
                    final.push_back(0);
                    return new symbolic_automaton(0, final, mvs1);
                } else {
                    return new symbolic_automaton();
                }
            }
            else {
                return new symbolic_automaton(0, final, mvs1);
            }
        }

        void add_to_final_states(unsigned s) {
            if (!is_final_state(s)) {
                m_final_set.insert(s);
                m_final_states.push_back(s);
            }
        }

        void remove_from_final_states(unsigned s) {
            if (is_final_state(s)) {
                m_final_set.erase(s);
                std::erase(m_final_states, s);
            }
        }

        bool is_sink_state(unsigned s) const {
            if (is_final_state(s)) return false;
            moves mvs;
            get_moves_from(s, mvs);
            for (move const& m : mvs) {
                if (s != m.dst()) return false;
            }
            return true;
        }

        void add_init_to_final_states() {
            add_to_final_states(init());
        }

        void add_final_to_init_moves() {
            for (unsigned i = 0; i < m_final_states.size(); ++i) {
                unsigned state = m_final_states[i];
                bool found = false;
                moves const& mvs = m_delta[state];
                for (move const &mv : mvs) {
                    found = (mv.dst() == m_init) && mv.is_epsilon();
                    if (found)
                        break;
                }
                if (!found && state != m_init) {
                    add(move(state, m_init));
                }
            }
        }

        // remove epsilon transitions
        // src - e -> dst
        // in_degree(src) = 1, final(src) => final(dst), src0 != src
        //    src0 - t -> src - e -> dst => src0 - t -> dst
        // out_degree(dst) = 1, final(dst) => final(src), dst != dst1
        //    src - e -> dst - t -> dst1 => src - t -> dst1

        // Generalized:
        // Src - E -> dst - t -> dst1 => Src - t -> dst1   if dst is final => each Src is final
        //
        // src - e -> dst - ET -> Dst1 => src - ET -> Dst1  if in_degree(dst) = 1, src != dst
        // Src - E -> dst - et -> dst1 => Src - et -> dst1  if out_degree(dst) = 1, src != dst
        //
        // Some missing:
        //  src - et -> dst - E -> Dst1 => src - et -> Dst1   if in_degree(dst) = 1
        //  Src - ET -> dst - e -> dst1 => Src - ET -> dst1  if out_degree(dst) = 1,
        //
        void compress() {
            assert(!m_delta.empty());
            moves to_remove;
            for (unsigned i = 0; i < m_delta.size(); ++i) {
                for (move const &mv : m_delta[i]) {
                    unsigned src = mv.src();
                    unsigned dst = mv.dst();
                    assert(src == i);
                    if (mv.is_epsilon()) {
                        if (src == dst) {
                            // just remove this edge.
                        }
                        else if (1 == in_degree(src) && 1 == out_degree(src) && init() != src && (!is_final_state(src) || is_final_state(dst))) {
                            move const &mv0 = *m_delta_inv[src].begin();
                            unsigned src0 = mv0.src();
                            std::optional<z3::expr const> const &t = mv0.t();
                            assert(mv0.dst() == src);
                            if (src0 == src) {
                                continue;
                            }
                            add(move(src0, dst, t));
                            to_remove.insert(move(src0, src, t));
                        }
                        else if (1 == out_degree(dst) && 1 == in_degree(dst) && init() != dst && (!is_final_state(dst) || is_final_state(src))) {
                            move const &mv1 = *m_delta[dst].begin();
                            unsigned dst1 = mv1.dst();
                            std::optional<z3::expr const> const &t = mv1.t();
                            assert(mv1.src() == dst);
                            if (dst1 == dst) {
                                continue;
                            }
                            add(move(src, dst1, t));
                            to_remove.insert(move(dst, dst1, t));
                        }
                        else if (1 == in_degree(dst) && (!is_final_state(dst) || is_final_state(src)) && init() != dst) {
                            moves const &mvs = m_delta[dst];
                            moves mvs1;
                            for (move const &mv : mvs) {
                                mvs1.insert(move(src, mv.dst(), mv.t()));
                            }
                            for (move const &mv : mvs1) {
                                to_remove.insert(move(dst, mv.dst(), mv.t()));
                                add(mv);
                            }
                        }
                        //
                        // Src - E -> dst - et -> dst1 => Src - et -> dst1  if out_degree(dst) = 1, src != dst
                        //
                        else if (1 == out_degree(dst) && all_epsilon_in(dst) && init() != dst && !is_final_state(dst)) {
                            move const& mv = *m_delta[dst].begin();
                            unsigned dst1 = mv.dst();
                            std::optional<z3::expr const> const &t = mv.t();
                            std::vector<unsigned> src0s;
                            moves const &mvs = m_delta_inv[dst];
                            moves mvs1;
                            for (move const &mv1 : mvs) {
                                assert(mv1.is_epsilon());
                                mvs1.insert(move(mv1.src(), dst1, t));
                            }
                            for (move const &mv1 : mvs1) {
                                to_remove.insert(move(mv1.src(), dst, std::nullopt));
                                add(mv1);
                            }
                            to_remove.insert(move(dst, dst1, t));
                            continue;
                        }
                        //
                        // Src1 - ET -> src - e -> dst => Src1 - ET -> dst  if out_degree(src) = 1, src != init()
                        //
                        else if (1 == out_degree(src) && init() != src && (!is_final_state(src) || is_final_state(dst))) {
                            moves const &mvs = m_delta_inv[src];
                            moves mvs1;
                            for (move const &mv : mvs) {
                                mvs1.insert(move(mv.src(), dst, mv.t()));
                            }
                            for (move const &mv : mvs1) {
                                to_remove.insert(move(mv.src(), src, mv.t()));
                                add(mv);
                            }
                        }
                        else if (1 == out_degree(src) && (is_final_state(src) || !is_final_state(dst))) {
                            moves const &mvs = m_delta[dst];
                            moves mvs1;
                            for (move const &mv : mvs) {
                                mvs1.insert(move(src, mv.dst(), mv.t()));
                            }
                            for (move const &mv : mvs1) {
                                add(mv);
                            }
                        }
                        else {
                            continue;
                        }
                        to_remove.insert(move(src, dst, std::nullopt));
                    }
                }
            }
            for (move const &m : to_remove)
                remove(m);
            assert(!m_delta.empty());
            while (true) {
                assert(!m_delta.empty());
                unsigned src = m_delta.size() - 1;
                if (in_degree(src) == 0 && init() != src) {
                    remove_from_final_states(src);
                    m_delta.pop_back();
                }
                else {
                    break;
                }
            }
            sinkify_dead_states();
        }

        bool is_sequence(unsigned &length) const {
            if (is_final_state(m_init) && (out_degree(m_init) == 0 || (out_degree(m_init) == 1 && is_loop_state(m_init)))) {
                length = 0;
                return true;
            }
            if (is_empty() || in_degree(m_init) != 0 || out_degree(m_init) != 1) {
                return false;
            }

            length = 1;
            unsigned s = get_move_from(m_init).dst();
            while (!is_final_state(s)) {
                if (out_degree(s) != 1 || in_degree(s) != 1) {
                    return false;
                }
                s = get_move_from(s).dst();
                ++length;
            }
            return out_degree(s) == 0 || (out_degree(s) == 1 && is_loop_state(s));
        }

        unsigned init() const { return m_init; }
        std::vector<unsigned> const& final_states() const { return m_final_states; }
        unsigned in_degree(unsigned state) const { return m_delta_inv[state].size(); }
        unsigned out_degree(unsigned state) const { return m_delta[state].size(); }
        move const& get_move_from(unsigned state) const { assert(m_delta[state].size() == 1); return *m_delta[state].begin(); }
        move const& get_move_to(unsigned state) const { assert(m_delta_inv[state].size() == 1); return *m_delta_inv[state].begin(); }
        moves const& get_moves_from(unsigned state) const { return m_delta[state]; }
        moves const& get_moves_to(unsigned state) const { return m_delta_inv[state]; }
        bool initial_state_is_source() const { return m_delta_inv[m_init].empty(); }
        bool is_final_state(unsigned s) const { return m_final_set.find(s) != m_final_set.end(); }
        bool is_final_configuration(std::set<unsigned> const &s) const {
            for (unsigned i : s) {
                if (is_final_state(i))
                    return true;
            }
            return false;
        }
        bool is_epsilon_free() const {
            for (moves const &mvs : m_delta) {
                for (move const &m : mvs) {
                    if (m.is_epsilon()) return false;
                }
            }
            return true;
        }

        bool all_epsilon_in(unsigned s) {
            moves const &mvs = m_delta_inv[s];
            for (move const &m : mvs) {
                if (!m.is_epsilon()) return false;
            }
            return true;
        }

        bool is_empty() const { return m_final_states.empty(); }
        bool is_epsilon() const { return m_final_states.size() == 1 && m_final_states.back() == init() && m_delta.empty(); }
        unsigned final_state() const { return m_final_states[0]; }
        bool has_single_final_sink() const { return m_final_states.size() == 1 && m_delta[final_state()].empty(); }
        unsigned num_states() const { return m_delta.size(); }
        bool is_loop_state(unsigned s) const {
            moves mvs;
            get_moves_from(s, mvs);
            for (move const &m : mvs) {
                if (s == m.dst()) return true;
            }
            return false;
        }

        /**
         * True if state is initial or has incoming or outgoing edges
         */
        bool is_state(unsigned s) const {
            if (s == init()) return true;
            if (s >= m_delta.size()) return false;
            return m_delta[s].size() > 0 || m_delta_inv[s].size() > 0;
        }

        /**
         * The largest state that *may* have transitions
         */
        unsigned max_state() const { return m_delta.size() - 1; }

        /**
         * Recalculates every time
         */
        unsigned num_moves() const {
            unsigned count = 0;
            for (int i = 0; i < m_delta.size(); i++)
                count += m_delta[i].size();
            return count;
        }

        unsigned move_count() const {
            unsigned result = 0;
            for (moves const& mvs : m_delta) result += mvs.size();
            return result;
        }
        void get_epsilon_closure(unsigned state, std::vector<unsigned>& states) const {
            get_epsilon_closure(state, m_delta, states);
        }
        void get_inv_epsilon_closure(unsigned state, std::vector<unsigned>& states) const {
            get_epsilon_closure(state, m_delta_inv, states);
        }
        void get_moves_from(unsigned state, moves& mvs, bool epsilon_closure = true) const {
            get_moves(state, m_delta, mvs, epsilon_closure);
        }
        void get_moves_from_states(std::set<unsigned> const& states, moves& mvs, bool epsilon_closure = true) const {
            for (unsigned i : states) {
                moves curr;
                get_moves(i, m_delta, curr, epsilon_closure);
                for(move const &mv : curr)
                    mvs.insert(mv);
            }
        }
        void get_moves_to(unsigned state, moves& mvs, bool epsilon_closure = true) {
            get_moves(state, m_delta_inv, mvs, epsilon_closure);
        }

        friend std::ostream& operator<<(
            std::ostream &output,
            symbolic_automaton const &f
        ) {
            f.display(output);
            return output;
        }

    private:

        std::ostream& display(std::ostream &out) const {
            out << "init: " << init() << "\n";
            out << "final: ";
            for (auto f : m_final_states)
                out << f << " ";
            out << "\n";

            for (unsigned i = 0; i < m_delta.size(); ++i) {
                moves const &mvs = m_delta[i];
                for (move const &mv : mvs) {
                    out << i << " -> " << mv.dst() << " ";
                    if (!mv.is_epsilon()) {
                        out << "if " << mv.t().value();
                    }
                    out << "\n";
                }
            }
            return out;
        }

        void sinkify_dead_states() {
            std::set<unsigned> dead_states;
            for (unsigned i = 0; i < m_delta.size(); ++i) {
                auto res = std::find(
                    begin(m_final_states),
                    end(m_final_states),
                    i
                );
                if (res == std::end(m_final_states)) {
                    dead_states.insert(i);
                }
            }
            bool change = true;
            std::vector<unsigned> to_remove;
            while (change) {
                change = false;
                to_remove.clear();
                for (unsigned s : dead_states) {
                    moves const& mvs = m_delta[s];
                    for (move const& mv : mvs) {
                        auto res = std::find(
                            begin(dead_states),
                            end(dead_states),
                            mv.dst()
                        );
                        if (res == std::end(dead_states)) {
                            to_remove.push_back(s);
                            break;
                        }
                    }
                }
                change = !to_remove.empty();
                for (unsigned s : to_remove) {
                    dead_states.erase(s);
                }
                to_remove.clear();
            }
            for (unsigned s : dead_states) {
                m_delta[s].clear();
            }
        }

        void add(move const &mv) {
            m_delta[mv.src()].insert(mv);
            m_delta_inv[mv.dst()].insert(mv);
        }

        void remove(move const &m, moves& mvs) {
            mvs.erase(m);
        }

        void remove(move const &m) {
            remove(m, m_delta[m.src()]);
            remove(m, m_delta_inv[m.dst()]);
        }

        mutable std::vector<unsigned> m_states1, m_states2;

        void get_moves(
            unsigned state,
            std::vector<moves> const &delta,
            moves &mvs,
            bool epsilon_closure
        ) const {
            m_states1.clear();
            m_states2.clear();
            get_epsilon_closure(state, delta, m_states1);
            for (unsigned i = 0; i < m_states1.size(); ++i) {
                state = m_states1[i];
                moves const &mv1 = delta[state];
                for (move const &mv : mv1) {
                    if (!mv.is_epsilon()) {
                        if (epsilon_closure) {
                            m_states2.clear();
                            get_epsilon_closure(mv.dst(), delta, m_states2);
                            for (unsigned k = 0; k < m_states2.size(); ++k) {
                                mvs.insert(move(state, m_states2[k], mv.t()));
                            }
                        }
                        else {
                            mvs.insert(move(state, mv.dst(), mv.t()));
                        }
                    }
                }
            }
        }

        void get_epsilon_closure(
            unsigned state,
            std::vector<moves> const &delta,
            std::vector<unsigned> &states
        ) const {
            m_todo.push_back(state);
            m_visited.insert(state);
            while (!m_todo.empty()) {
                state = m_todo.back();
                states.push_back(state);
                m_todo.pop_back();
                moves const &mvs = delta[state];
                for (move const &mv : mvs) {
                    unsigned tgt = mv.dst();
                    if (
                        mv.is_epsilon() &&
                        m_visited.find(tgt) == m_visited.end()
                    ) {
                        m_visited.insert(tgt);
                        m_todo.push_back(tgt);
                    }
                }
            }
            m_visited.clear();
            assert(m_todo.empty());
        }

        static void append_moves(
            unsigned offset,
            symbolic_automaton const &a,
            moves &mvs
        ) {
            for (unsigned i = 0; i < a.num_states(); ++i) {
                moves const &mvs1 = a.m_delta[i];
                for (move const &mv : mvs1) {
                    mvs.insert(move(mv.src() + offset, mv.dst() + offset, mv.t()));
                }
            }
        }

        static void append_final(
            unsigned offset,
            symbolic_automaton const &a,
            std::vector<unsigned> &final
        ) {
            for (unsigned s : a.m_final_states) {
                final.push_back(s+offset);
            }
        }

    private:

        /**
         * First returned is map from pred id to sat or not
         */
        std::vector<std::pair<std::map<unsigned, bool>, z3::expr>>
        generate_min_terms(
            z3::context &context, std::vector<z3::expr> &constraints
        ) const {
            std::vector<
                std::pair<std::map<unsigned, bool>, z3::expr>
            > min_terms;

            z3::expr curr_pred = context.bool_val(true);
            std::map<unsigned, bool> curr_bv;

            z3::solver solver(context);

            generate_min_terms_rec(
                context,
                solver,
                constraints,
                min_terms,
                0,
                curr_bv,
                curr_pred
            );

            return min_terms;
        }

        void generate_min_terms_rec(
            z3::context &context,
            z3::solver &solver,
            std::vector<z3::expr> &constraints,
            std::vector<
                std::pair<std::map<unsigned, bool>, z3::expr>
            > &min_terms,
            unsigned i,
            std::map<unsigned, bool> &curr_bv,
            z3::expr &curr_pred
        ) const {
            solver.push();
            solver.add(curr_pred);
            auto is_sat = solver.check();
            solver.pop();

            assert(is_sat != z3::unknown);
            // ^ incomplete theory: unable to generate min-terms");

            if (is_sat != z3::sat) {
                return;
            }

            if (i == constraints.size()) {
                min_terms.push_back(
                    std::pair<std::map<unsigned, bool>, z3::expr>(
                        curr_bv,
                        curr_pred
                    )
                );
            } else {
                //true case
                curr_bv.insert_or_assign(constraints[i].id(), true);
                z3::expr new_pred_pos = curr_pred && constraints[i];
                generate_min_terms_rec(
                    context,
                    solver,
                    constraints,
                    min_terms,
                    i + 1,
                    curr_bv,
                    new_pred_pos
                );

                curr_bv.insert_or_assign(constraints[i].id(), false);

                z3::expr neg = !constraints[i];
                z3::expr new_pred_neg = curr_pred && neg;
                generate_min_terms_rec(
                    context,
                    solver,
                    constraints,
                    min_terms,
                    i + 1,
                    curr_bv,
                    new_pred_neg
                );

                curr_bv.erase(constraints[i].id());
            }
        }

        static bool move_t_eq(
            std::optional<z3::expr const> const &t1,
            std::optional<z3::expr const> const &t2
        ) {
            return (!t1.has_value() && !t2.has_value()) ||
                (t1.has_value() && t2.has_value()
                    && t1.value().id() == t2.value().id());
        }

        static bool move_t_lt(
            std::optional<z3::expr const> const &t1,
            std::optional<z3::expr const> const &t2
        ) {
            if (!t1.has_value() && !t2.has_value())
                return false;
            if (!t1.has_value() && t2.has_value())
                return true;
            if (t1.has_value() && !t2.has_value())
                return false;
            return t1.value().id() < t2.value().id();
        }
    };
}
#endif
