
#include <iostream>

#include <catch2/catch_test_macros.hpp>

#include "z3++.h"

#include "parikh_image_provider.h"
#include "re2automaton.h"
#include "symbolic_automaton.h"
#include "util.h"

using namespace std;
using namespace sym_parikh;
using namespace z3;

static ParikhImageProviderOne pi_provider1;
static ParikhImageProviderTwo pi_provider2;

TEST_CASE("Single char aut image", "[parikh-single-char]") {
    context c;
    expr re = smtlib2re(c, "(re.range \"a\" \"a\")");

    vector<expr> predicates;
    vector<expr> pred_vars;

    expr fv = re2automaton::get_free_variable(c);

    predicates.push_back(fv == c.int_val((int) 'a'));
    expr fa = c.int_const("fa");
    pred_vars.push_back(fa);

    auto pi = pi_provider1.get_parikh_image(c, re, predicates, pred_vars);

    auto image = pi->get_image_formula();

    solver s(c);
    s.add(image);

    s.push();
    s.add(fa == 0);
    REQUIRE(s.check() == unsat);
    s.pop();

    s.push();
    s.add(fa == 1);
    REQUIRE(s.check() == sat);
    s.pop();

    s.push();
    s.add(fa == 2);
    REQUIRE(s.check() == unsat);
    s.pop();
}

TEST_CASE("Single hex char aut image", "[parikh-single-char]") {
    context c;
    expr re = smtlib2re(c, "(str.to_re (_ char #x1a))");

    vector<expr> predicates;
    vector<expr> pred_vars;

    expr fv = re2automaton::get_free_variable(c);

    predicates.push_back(fv == c.int_val(26));
    expr fa = c.int_const("fa");
    pred_vars.push_back(fa);

    auto pi = pi_provider1.get_parikh_image(c, re, predicates, pred_vars);

    auto image = pi->get_image_formula();

    solver s(c);
    s.add(image);

    s.push();
    s.add(fa == 0);
    REQUIRE(s.check() == unsat);
    s.pop();

    s.push();
    s.add(fa == 1);
    REQUIRE(s.check() == sat);
    s.pop();

    s.push();
    s.add(fa == 2);
    REQUIRE(s.check() == unsat);
    s.pop();
}

TEST_CASE("Char loop with pred count one", "[parikh-loop-preds-one]") {
    context c;
    expr re = smtlib2re(c,
        "(re.++ "
            "(re.range \"a\" \"b\")"
            "(re.++ "
                "(re.* (re.range \"b\" \"c\")) "
                "(re.range \"a\" \"c\")"
            ")"
        ")"
    );

    expr fv = re2automaton::get_free_variable(c);

    vector<expr> predicates;
    vector<expr> pred_vars;

    predicates.push_back(fv == c.int_val((int) 'a'));
    pred_vars.push_back(c.int_const("fa"));
    predicates.push_back(fv == c.int_val((int) 'c'));
    pred_vars.push_back(c.int_const("fc"));

    auto pi = pi_provider1.get_parikh_image(c, re, predicates, pred_vars);

    auto image = pi->get_image_formula();

    solver s(c);
    s.add(image);

    s.push();
    s.add(pred_vars[0] == 1);
    s.add(pred_vars[1] == 5);
    REQUIRE(s.check() == sat);
    s.pop();

    s.push();
    s.add(pred_vars[0] == 5);
    s.add(pred_vars[1] == 5);
    REQUIRE(s.check() == unsat);
    s.pop();

    s.push();
    s.add(pred_vars[0] == 2);
    s.add(pred_vars[1] == 5);
    REQUIRE(s.check() == sat);
    s.pop();
}

// TODO: make parent class of ParikhImage and avoid repeated test code
TEST_CASE("Char loop with pred count two", "[parikh-loop-preds-two]") {
    context c;
    expr re = smtlib2re(c,
        "(re.++ "
            "(re.range \"a\" \"b\")"
            "(re.++ "
                "(re.* (re.range \"b\" \"c\")) "
                "(re.range \"a\" \"c\")"
            ")"
        ")"
    );

    expr fv = re2automaton::get_free_variable(c);

    vector<expr> predicates;
    vector<expr> pred_vars;

    predicates.push_back(fv == c.int_val((int) 'a'));
    pred_vars.push_back(c.int_const("fa"));
    predicates.push_back(fv == c.int_val((int) 'c'));
    pred_vars.push_back(c.int_const("fc"));

    auto pi = pi_provider2.get_parikh_image(c, re, predicates, pred_vars);

    auto image = pi->get_image_formula();

    solver s(c);
    s.add(image);

    s.push();
    s.add(pred_vars[0] == 1);
    s.add(pred_vars[1] == 5);
    REQUIRE(s.check() == sat);
    s.pop();

    s.push();
    s.add(pred_vars[0] == 5);
    s.add(pred_vars[1] == 5);
    REQUIRE(s.check() == unsat);
    s.pop();

    s.push();
    s.add(pred_vars[0] == 2);
    s.add(pred_vars[1] == 5);
    REQUIRE(s.check() == sat);
    s.pop();
}

TEST_CASE("Negated astar", "[neg-astar]") {
    context c;
    expr re = smtlib2re(c, "(re.comp (re.* (str.to_re \"a\")))");

    expr fv = re2automaton::get_free_variable(c);

    vector<expr> predicates;
    vector<expr> pred_vars;

    predicates.push_back(fv != c.int_val((int) 'a'));
    pred_vars.push_back(c.int_const("fa"));

    auto pi = pi_provider1.get_parikh_image(c, re, predicates, pred_vars);

    auto image = pi->get_image_formula();

    solver s(c);
    s.add(image);

    s.push();
    s.add(pred_vars[0] == 1);
    REQUIRE(s.check() == sat);
    s.pop();

    s.push();
    s.add(pred_vars[0] == 0);
    REQUIRE(s.check() == unsat);
    s.pop();
}

