
#include <iostream>

#include <catch2/catch_test_macros.hpp>

#include "z3++.h"

#include "re2automaton.h"
#include "symbolic_automaton.h"
#include "util.h"

using namespace std;
using namespace sym_parikh;
using namespace z3;

TEST_CASE("Single char regex to automaton", "[re2automaton-single-char]") {
    context c;
    re2automaton re2aut(c);

    expr re = smtlib2re(c, "(re.range \"a\" \"b\")");
    shared_ptr<symbolic_automaton> aut = re2aut.translate(re);

    REQUIRE(aut->num_states() == 2);

    symbolic_automaton::moves const &moves = aut->get_moves_from(aut->init());
    REQUIRE(moves.size() == 1);

    symbolic_automaton::move const &m = *moves.begin();
    REQUIRE(aut->is_final_state(m.dst()));

    // check pred is (str.in_re x (re.range "a" "b"))
    expr pred = re2aut.get_in_range('a', 'b');
    REQUIRE(m.t().value().id() == pred.id());

    REQUIRE(aut->get_moves_from(m.dst()).size() == 0);
}

TEST_CASE("Seq regex to automaton", "[re2automaton-seq]") {
    context c;
    re2automaton re2aut(c);
    string seq("abc");

    expr re = smtlib2re(c, "(str.to_re \"" + seq + "\")");
    shared_ptr<symbolic_automaton> aut = re2aut.translate(re);

    REQUIRE(aut->num_states() == 4);

    unsigned cur_state = aut->init();
    for (char chr : seq) {
        symbolic_automaton::moves const &moves =
            aut->get_moves_from(cur_state);
        REQUIRE(moves.size() == 1);

        // check pred is x == chr
        symbolic_automaton::move const &m = *moves.begin();
        expr pred = re2aut.get_is_char(chr);
        REQUIRE(m.t().value().id() == pred.id());

        cur_state = m.dst();
    }

    REQUIRE(aut->is_final_state(cur_state));
}

TEST_CASE("Intersection regex to automaton", "[re2automaton-intersection]") {
    context c;
    re2automaton re2aut(c);

    expr re = smtlib2re(
        c,
        "(re.inter "
          "(re.range \"a\" \"c\")"
          "(re.range \"b\" \"d\")"
        ")"
    );
    shared_ptr<symbolic_automaton> aut = re2aut.translate(re);

    REQUIRE(aut->num_states() == 2);

    symbolic_automaton::moves const &moves = aut->get_moves_from(aut->init());
    REQUIRE(moves.size() == 1);

    // check pred is intersection
    symbolic_automaton::move const &m = *moves.begin();
    expr pred = re2aut.get_in_range('a', 'c') && re2aut.get_in_range('b', 'd');

    REQUIRE(m.t().value().id() == pred.id());

    REQUIRE(aut->is_final_state(m.dst()));
}

TEST_CASE("Union regex to automaton and all char", "[re2automaton-union]") {
    context c;
    re2automaton re2aut(c);

    expr re = smtlib2re(
        c,
        "(re.union"
          "(re.range \"a\" \"c\")"
          "re.allchar"
        ")"
    );
    shared_ptr<symbolic_automaton> aut = re2aut.translate(re);

    symbolic_automaton::moves const &moves = aut->get_moves_from(aut->init());
    REQUIRE(moves.size() == 2);

    // check first pred is range a c
    symbolic_automaton::move const &m1 = *moves.begin();
    expr pred1 = re2aut.get_in_range('a', 'c');
    REQUIRE(m1.t().value().id() == pred1.id());
    REQUIRE(aut->is_final_state(m1.dst()));

    // check second pred is *
    symbolic_automaton::move const &m2 = *(++moves.begin());
    expr pred2 = c.bool_val(true);
    REQUIRE(m2.t().value().id() == pred2.id());
    REQUIRE(aut->is_final_state(m2.dst()));
}

TEST_CASE("Star regex to automaton", "[re2automaton-star]") {
    context c;
    re2automaton re2aut(c);

    expr re = smtlib2re(
        c,
        "(re.*"
          "(re.range \"a\" \"c\")"
        ")"
    );
    shared_ptr<symbolic_automaton> aut = re2aut.translate(re);

    symbolic_automaton::moves const &moves = aut->get_moves_from(aut->init());
    REQUIRE(moves.size() == 1);

    // check tran has pred is range a c and is loop
    symbolic_automaton::move const &m1 = *moves.begin();
    expr pred1 = re2aut.get_in_range('a', 'c');
    REQUIRE(m1.t().value().id() == pred1.id());
    REQUIRE(m1.dst() == aut->init());
    REQUIRE(aut->is_final_state(m1.dst()));
}

